from kawin.thermo.Thermodynamics import GeneralThermodynamics
import numpy as np
from pycalphad import variables as v
from kawin.thermo.Mobility import inverseMobility, inverseMobility_from_diffusivity, tracer_diffusivity, tracer_diffusivity_from_diff
from kawin.thermo.FreeEnergyHessian import dMudX
from kawin.thermo.LocalEquilibrium import local_equilibrium

class MulticomponentThermodynamics (GeneralThermodynamics):
    '''
    Class for defining driving force and (possibly) interfacial composition functions
    for a multicomponent system using pyCalphad and thermodynamic databases

    Parameters
    ----------
    database : str
        File name for database
    elements : list
        Elements to consider
        Note: reference element must be the first index in the list
    phases : list
        Phases involved
        Note: matrix phase must be first index in the list
    drivingForceMethod : str (optional)
        Method used to calculate driving force
        Options are 'tangent' (default), 'approximate', 'sampling' and 'curvature' (not recommended)
    parameters : list [str] or dict {str : float}
        List of parameters to keep symbolic in the thermodynamic or mobility models
    '''
    def __init__(self, database, elements, phases, drivingForceMethod = 'tangent', parameters = None):
        super().__init__(database, elements, phases, drivingForceMethod, parameters)

        #Previous variables for curvature terms
        #Near saturation, pycalphad may detect only a single phase (if sampling density is too low)
        #When this occurs, this will assume that the system is on the same tie-line and
        #use the previously calculated values
        self._prevDc = {p: None for p in phases[1:]}
        self._prevMc = {p: None for p in phases[1:]}
        self._prevGba = {p: None for p in phases[1:]}
        self._prevBeta = {p: None for p in phases[1:]}
        self._prevCa = {p: None for p in phases[1:]}
        self._prevCb = {p: None for p in phases[1:]}

    def getInterfacialComposition(self, x, T, gExtra = 0, precPhase = None):
        '''
        Gets interfacial composition by calculating equilibrum with Gibbs-Thomson effect

        Parameters
        ----------
        T : float or array
            Temperature in K
        gExtra : float or array (optional)
            Extra contributions to the precipitate Gibbs free energy
            Gibbs Thomson contribution defined as Vm * (2*gamma/R + g_Elastic)
            Defaults to 0
        precPhase : str
            Precipitate phase to consider (default is first precipitate in list)

        Note: for multiple conditions, only gExtra has to be an array
            This will calculate compositions for multiple gExtra at the input Temperature

            If T is also an array, then T and gExtra must be the same length
            where each index will pertain to a single condition

        Returns
        -------
        (parent composition, precipitate composition)
        Both will be either float or array based off shape of gExtra
        Will return (None, None) if precipitate is unstable
        '''
        if hasattr(gExtra, '__len__'):
            if not hasattr(T, '__len__'):
                T = T * np.ones(len(gExtra))

            caArray = []
            cbArray = []
            for i in range(len(gExtra)):
                ca, cb = self._interfacialComposition(x, T[i], gExtra[i], precPhase)
                caArray.append(ca)
                cbArray.append(cb)
            caArray = np.array(caArray)
            cbArray = np.array(cbArray)
            return caArray, cbArray
        else:
            return self._interfacialComposition(x, T, gExtra, precPhase)


    def _interfacialComposition(self, x, T, gExtra = 0, precPhase = None):
        '''
        Gets interfacial composition, will return None, None if composition is in single phase region

        Parameters
        ----------
        T : float
            Temperature in K
        gExtra : float (optional)
            Extra contributions to the precipitate Gibbs free energy
            Gibbs Thomson contribution defined as Vm * (2*gamma/R + g_Elastic)
            Defaults to 0
        precPhase : str
            Precipitate phase to consider (default is first precipitate in list)

        Returns
        -------
        (parent composition, precipitate composition)
        Both will be either float or array based off shape of gExtra
        Will return (None, None) if precipitate is unstable
        '''
        if precPhase is None:
            precPhase = self.phases[1]

        wks = self.getEq(x, T, gExtra, precPhase)
        mu = np.squeeze(wks.eq.MU)

        #Check for convergence, return None if not converged
        if np.any(np.isnan(mu)):
            return None, None
        
        cs_list = wks.get_composition_sets()
        ph = [cs.phase_record.phase_name for cs in cs_list]

        #Check if matrix and precipitate phase are stable, and check if there's no miscibility gaps
        if len(ph) == 2 and self.phases[0] in ph and precPhase in ph:
            sortIndices = np.argsort(self.elements[:-1])
            unsortIndices = np.argsort(sortIndices)

            mu = mu[unsortIndices]

            cs_matrix = cs_list[ph.index(self.phases[0])]
            xM = np.array(cs_matrix.X, dtype=np.float64)
            xM = xM[unsortIndices]

            cs_precip = cs_list[ph.index(precPhase)]
            xP = np.array(cs_precip.X, dtype=np.float64)
            xP = xP[unsortIndices]

            return xM, xP

        return None, None

    def _curvatureFactorFromEq(self, chemical_potentials, composition_sets, precPhase=None):
        '''
        Curvature factor (from Phillipes and Voorhees - 2013)

        Steps
            1. Check that there is 2 phases in equilibrium, one being the matrix and the other being precipitate
            2. Get Dnkj, dmu/dx and inverse mobility term from composition set of matrix phase
            3. Get dmu/dx of precipitate phase
            4. Get difference in matrix and precipitate phase composition (we use a second order approximation to get precipitate composition as function of R)
            5. Compute numerator, denominator, Gba and beta term
                Denominator (X_bar^T * invMob * X_bar) is used for growth rate (eq 28)
                Numerator (D^-1 * X_bar), denominator is used for matrix interfacial composition (eq 31)
                Gba and matrix interfacial composition is used for precipitate interfacial composition (eq 36)
                    Gba here is (dmu/dx_beta)^-1 * dmu/dx_alpha
                Note: these equations have a term X_bar^T * dmu/dx_alpha * X_bar_infty, but this is just the driving force so we don't need to calculate it here

        Parameters
        ----------
        chemical_potentials : 1-D float64 array
        composition_sets : List[pycalphad.composition_set.CompositionSet]
        precPhase : str (optional)
            Precipitate phase (defaults to first precipitate in list)

        Returns
        -------
        {D-1 dCbar / dCbar^T M-1 dCbar} - for calculating interfacial composition of matrix
        {1 / dCbar^T M-1 dCbar} - for calculating growth rate
        {Gb^-1 Ga} - for calculating precipitate composition
        beta - Impingement rate
        Ca - interfacial composition of matrix phase
        Cb - interfacial composition of precipitate phase

        Will return (None, None, None, None, None, None) if single phase
        '''
        if precPhase is None:
            precPhase = self.phases[1]

        ele = list(composition_sets[0].phase_record.nonvacant_elements)
        refIndex = ele.index(self.elements[0])

        ph = [cs.phase_record.phase_name for cs in composition_sets]

        if len(ph) == 2 and self.phases[0] in ph and precPhase in ph:
            sortIndices = np.argsort(self.elements[1:-1])
            unsortIndices = np.argsort(sortIndices)

            matrix_cs = [cs for cs in composition_sets if cs.phase_record.phase_name == self.phases[0]][0]

            if self.mobCallables[self.phases[0]] is None:
                Dnkj, dMudxParent, invMob = inverseMobility_from_diffusivity(chemical_potentials, matrix_cs,
                                                                             self.elements[0], self.diffCallables[self.phases[0]],
                                                                             diffusivity_correction=self.mobility_correction, parameters=self._parameters)

                #NOTE: This is note tested yet
                Dtrace = tracer_diffusivity_from_diff(matrix_cs, self.diffCallables[self.phases[0]], diffusivity_correction=self.mobility_correction, parameters=self._parameters)
            else:
                Dnkj, dMudxParent, invMob = inverseMobility(chemical_potentials, matrix_cs, self.elements[0],
                                                            self.mobCallables[self.phases[0]],
                                                            mobility_correction=self.mobility_correction, parameters=self._parameters)
                Dtrace = tracer_diffusivity(matrix_cs, self.mobCallables[self.phases[0]], mobility_correction=self.mobility_correction, parameters=self._parameters)

            xMFull = np.array(matrix_cs.X)
            xM = np.delete(xMFull, refIndex)

            precip_cs = [cs for cs in composition_sets if cs.phase_record.phase_name == precPhase][0]
            dMudxPrec = dMudX(chemical_potentials, precip_cs, self.elements[0])
            xPFull = np.array(precip_cs.X)
            xP = np.delete(xPFull, refIndex)
            xBarFull = np.array([xPFull - xMFull])
            xBar = np.array([xP - xM])

            num = np.matmul(np.linalg.inv(Dnkj), xBar.T).flatten()

            #Denominator should be a scalar since its V * M * V^T
            den = np.matmul(xBar, np.matmul(invMob, xBar.T)).flatten()[0]

            if np.linalg.matrix_rank(dMudxPrec) == dMudxPrec.shape[0]:
                Gba = np.matmul(np.linalg.inv(dMudxPrec), dMudxParent)
                Gba = Gba[unsortIndices,:]
                Gba = Gba[:,unsortIndices]
            else:
                Gba = np.zeros(dMudxPrec.shape)

            betaNum = xBarFull**2
            betaDen = Dtrace * xMFull.flatten()
            bsum = np.sum(betaNum / betaDen)
            if bsum == 0:
                beta = self._prevBeta[precPhase]
            else:
                beta = 1 / bsum

            self._prevDc[precPhase] = num[unsortIndices] / den
            self._prevMc[precPhase] = 1 / den
            self._prevGba[precPhase] = Gba
            self._prevBeta[precPhase] = beta
            self._prevCa[precPhase] = xM[unsortIndices]
            self._prevCb[precPhase] = xP[unsortIndices]

            return self._prevDc[precPhase], self._prevMc[precPhase], self._prevGba[precPhase], self._prevBeta[precPhase], self._prevCa[precPhase], self._prevCb[precPhase]
        else:
            return None

    def curvatureFactor(self, x, T, precPhase = None, training = False, searchDir = None):
        '''
        Curvature factor (from Phillipes and Voorhees - 2013) from composition and temperature
        This is the same as curvatureFactorEq, but will calculate equilibrium from x and T first

        Parameters
        ----------
        x : array
            Composition of solutes
        T : float
            Temperature
        precPhase : str (optional)
            Precipitate phase (defaults to first precipitate in list)
        searchDir : None or array
            If two-phase equilibrium is not present, then move x towards this composition to find two-phase equilibria
        training : bool (optional)
            If True, this will not cache any equilibrium
            This is used for training since training points may not be near each other

        Returns
        -------
        {D-1 dCbar / dCbar^T M-1 dCbar} - for calculating interfacial composition of matrix
        {1 / dCbar^T M-1 dCbar} - for calculating growth rate
        {Gb^-1 Ga} - for calculating precipitate composition
        beta - Impingement rate
        Ca - interfacial composition of matrix phase
        Cb - interfacial composition of precipitate phase

        Will return (None, None, None, None, None, None) if single phase
        '''
        if precPhase is None:
            precPhase = self.phases[1]
        if not hasattr(x, '__len__'):
            x = [x]

        #Remove first element if x lists composition of all elements
        if len(x) == len(self.elements) - 1:
            x = x[1:]
        cond = self._getConditions(x, T, 0)

        #Perform equilibrium from scratch if cache not set or when training surrogate
        if self._compset_cache.get(precPhase, None) is None or training:
            cs_results = self._getCompositionSetsForCurvature(x, T, precPhase)
            if cs_results is None:
                return None
            
            chemical_potentials, composition_sets = cs_results
        else:
            phases, sub_models = self._setupSubModels(precPhase)
            result, composition_sets = local_equilibrium(self.db, self.elements, phases, cond,
                                                         sub_models, self.phase_records,
                                                         composition_sets=self._compset_cache[precPhase])
            self._compset_cache[precPhase] = composition_sets
            chemical_potentials = result.chemical_potentials

        #Check if input equilibrium has converged
        if np.any(np.isnan(chemical_potentials)):
            if training:
                return None
            else:
                print('Warning: equilibrum was not able to be solved for, using results of previous calculation')
                return self._prevDc[precPhase], self._prevMc[precPhase], self._prevGba[precPhase], self._prevBeta[precPhase], self._prevCa[precPhase], self._prevCb[precPhase]

        ph = [cs.phase_record.phase_name for cs in composition_sets]
        if len(ph) == 2 and self.phases[0] in ph and precPhase in ph:
            return self._curvatureFactorFromEq(chemical_potentials, composition_sets, precPhase)
        #If in a singl phase region, we want to go along a search direction to find the nearest two phase region
        #    We then use this two-phase region to calculate growth rate (which should all be negative for dissolution)
        #    In PrecipitateModel, searchDir is the previous precipitate nucleate composition
        #    We performe a rouch search 
        elif searchDir is not None:
            currX = np.array(x)
            searchDir = np.array(searchDir)
            currX = 0.5 * currX + 0.5 * searchDir
            foundTwoPhases = False
            maxIt = 15
            currIt = 0
            while not foundTwoPhases:
                cs_results = self._getCompositionSetsForCurvature(currX, T, precPhase)
                if cs_results is None:
                    return None
                chemical_potentials, composition_sets = cs_results
                ph = [cs.phase_record.phase_name for cs in composition_sets]
                if len(ph) == 2 and self.phases[0] in ph and precPhase in ph:
                    foundTwoPhases = True
                elif len(ph) == 1 and self.phases[0] in ph:
                    #Only matrix is stable, move closer to searchDir
                    currX = 0.5*currX + 0.5*searchDir
                elif len(ph) == 1 and precPhase in ph:
                    #Only precipitate is stable, move closer to original x
                    currX = 0.5*currX + 0.5*np.array(x)

                #More than likely, this is not needed, but just in case
                #MaxIt is 15, which refers to a 6e-5 difference in test composition between the 14th and 15th iteration
                #    Which is probably more than enough to find a two-phase region
                currIt += 1
                if currIt > maxIt:
                    return None
            
            chemical_potentials, composition_sets = cs_results
            return self._curvatureFactorFromEq(chemical_potentials, composition_sets, precPhase)
        else:
            return None

    def getGrowthAndInterfacialComposition(self, x, T, dG, R, gExtra, precPhase = None, training = False, searchDir = None):
        '''
        Returns growth rate and interfacial compostion given Gibbs-Thomson contribution

        Parameters
        ----------
        x : array
            Composition of solutes
        T : float
            Temperature
        dG : float
            Driving force at given x and T
        R : float or array
            Precipitate radius
        gExtra : float or array
            Gibbs-Thomson contribution (must be same shape as R)
        precPhase : str (optional)
            Precipitate phase (defaults to first precipitate in list)
        searchDir : None or array
            If two-phase equilibrium is not present, then move x towards this composition to find a two-phase region
        training : bool (optional)
            If True, this will not cache any equilibrium
            This is used for training since training points may not be near each other

        Returns
        -------
        (growth rate, matrix composition, precipitate composition, equilibrium matrix comp, equilibrium precipitate comp)
        growth rate will be float or array based off shape of R
        matrix and precipitate composition will be array or 2D array based
            off shape of R
        '''
        if hasattr(R, '__len__'):
            R = np.array(R)
        if hasattr(gExtra, '__len__'):
            gExtra = np.array(gExtra)

        curv_results = self.curvatureFactor(x, T, precPhase, training, searchDir)
        if curv_results is None:
            return None, None, None, None, None
    
        dc, mc, gba, beta, ca, cb = curv_results

        #dc, mc, gba, beta, ca, cb = self.curvatureFactor(x, T, precPhase, training, searchDir)
        #if dc is None:
        #    return None, None, None, None, None

        Rdiff = (dG - gExtra)

        gr = (mc / R) * Rdiff

        if hasattr(Rdiff, '__len__'):
            calpha = np.zeros((len(Rdiff), len(self.elements[1:-1])))
            dca = np.zeros((len(Rdiff), len(self.elements[1:-1])))
            cbeta = np.zeros((len(Rdiff), len(self.elements[1:-1])))
            for i in range(len(self.elements[1:-1])):
                calpha[:,i] = x[i] - dc[i] * Rdiff
                dca[:,i] = calpha[:,i] - ca[i]

            dcb = np.matmul(gba, dca.T)
            for i in range(len(self.elements[1:-1])):
                cbeta[:,i] = cb[i] + dcb[i,:]

            calpha[calpha < 0] = 0
            calpha[calpha > 1] = 1
            cbeta[cbeta < 0] = 0
            cbeta[cbeta > 1] = 1

            return gr, calpha, cbeta, ca, cb
        else:
            calpha = x - dc * Rdiff
            cbeta = cb + np.matmul(gba, (calpha - ca)).flatten()

            calpha[calpha < 0] = 0
            calpha[calpha > 1] = 1
            cbeta[cbeta < 0] = 0
            cbeta[cbeta > 1] = 1

            return gr, calpha, cbeta, ca, cb

    def impingementFactor(self, x, T, precPhase = None, training = False):
        '''
        Returns impingement factor for nucleation rate calculations

        Parameters
        ----------
        x : array
            Composition of solutes
        T : float
            Temperature
        precPhase : str (optional)
            Precipitate phase (defaults to first precipitate in list)
        training : bool (optional)
            If True, this will not cache any equilibrium
            This is used for training since training points may not be near each other
        '''
        curv_results = self.curvatureFactor(x, T, precPhase, training)
        if curv_results is None:
            return self._prevBeta[precPhase]
        dc, mc, gba, beta, ca, cb = curv_results
        return beta
    
    def _getCompositionSetsForCurvature(self, x, T, precPhase):
        '''
        Create composition sets from equilibrium to be used for curvature factor

        Parameters
        ----------
        x : array
            Composition of solutes
        T : float
            Temperature
        precPhase : str (optional)
            Precipitate phase (defaults to first precipitate in list)
        '''
        cond = self._getConditions(x, T, 0)
        wks = self.getEq(x, T, 0, precPhase)
        chemical_potentials = np.array(wks.eq.MU, dtype=np.float64)
        composition_sets = wks.get_composition_sets()

        cs_list_matrix = [cs for cs in composition_sets if cs.phase_record.phase_name == self.phases[0]]
        cs_list_precip = [cs for cs in composition_sets if cs.phase_record.phase_name == precPhase]
        if len(cs_list_matrix) == 0:
            return None
        
        if len(cs_list_precip) == 0:
            composition_sets = [cs_list_matrix[0]]
            self._compset_cache[precPhase] = None
        else:
            miscMatrix = len(cs_list_matrix) > 1
            miscPrec = len(cs_list_precip) > 1
            self._compset_cache[precPhase] = composition_sets

            if miscMatrix or miscPrec:
                phases, sub_models = self._setupSubModels(precPhase)
                result, composition_sets = local_equilibrium(self.db, self.elements, phases, cond,
                                                        sub_models, self.phase_records,
                                                        composition_sets=self._compset_cache[precPhase])
                self._compset_cache[precPhase] = composition_sets
                chemical_potentials = result.chemical_potentials

        return chemical_potentials, composition_sets
    
    def _curvatureWithSearch(self, x, T, precPhase = None, training = True):
        '''
        Performs driving force calculation to get xb, which can be used to find
        curvature factors when driving force is negative. Main use is for the surrogate model
        to train on all points

        Parameters
        ----------
        x : array
            Composition of solutes
        T : float
            Temperature
        precPhase : str (optional)
            Precipitate phase (defaults to first precipitate in list)
        training : bool (optional)
            If True, this will not cache any equilibrium
            This is used for training since training points may not be near each other
        '''
        dg, xb = self.getDrivingForce(x, T, precPhase, returnComp = True, training = training)
        return self.curvatureFactor(x, T, precPhase, training = training, searchDir=xb)


import numpy as np
from kawin.precipitation.non_ideal.EffectiveDiffusion import EffectiveDiffusionFunctions
from kawin.precipitation.non_ideal.ShapeFactors import ShapeFactor
from kawin.precipitation.non_ideal.ElasticFactors import StrainEnergy
from kawin.precipitation.non_ideal.GrainBoundaries import GBFactors
from kawin.GenericModel import GenericModel
from enum import Enum

class VolumeParameter(Enum):
    MOLAR_VOLUME = 0
    ATOMIC_VOLUME = 1
    LATTICE_PARAMETER = 2

class PrecipitateBase(GenericModel):
    '''
    Base class for precipitation models

    Parameters
    ----------
    phases : list (optional)
        Precipitate phases (array of str)
        If only one phase is considered, the default is ['beta']
    elements : list (optional)
        Solute elements in system
        Note: order of elements must correspond to order of elements set in Thermodynamics module
                Also, the list here should just be the solutes while the Thermodynamics module needs also the parent element
        If binary system, then defualt is ['solute']
    '''
    def __init__(self, phases = ['beta'], elements = ['solute']):
        super().__init__()
        self.elements = elements
        self.numberOfElements = len(elements)
        self.phases = np.array(phases)

        self._resetArrays()
        self.resetConstraints()
        self._isSetup = False
        self._currY = None

        #Constants
        self.Rg = 8.314     #Gas constant - J/mol-K
        self.avo = 6.022e23 #Avogadro's number (/mol)
        self.kB = self.Rg / self.avo    #Boltzmann constant (J/K)

        #Default variables, these terms won't have to be set before simulation
        self.strainEnergy = [StrainEnergy() for i in self.phases]
        self.calculateAspectRatio = [False for i in self.phases]
        self.RdrivingForceLimit = np.zeros(len(self.phases), dtype=np.float32)
        self.shapeFactors = [ShapeFactor() for i in self.phases]
        self.theta = 2 * np.ones(len(self.phases), dtype=np.float32)
        self.effDiffFuncs = EffectiveDiffusionFunctions()
        self.effDiffDistance = self.effDiffFuncs.effectiveDiffusionDistance
        self.infinitePrecipitateDiffusion = [True for i in self.phases]
        self.dTemp = 0
        self.iterationSinceTempChange = 0
        self.GBenergy = 0.3     #J/m2
        self.parentPhases = [[] for i in self.phases]
        self.GB = [GBFactors() for p in self.phases]
        
        #Set other variables to None to throw errors if not set
        self.xInit = None
        self.Tparameters = None

        #Nucleation site density, it will default to dislocations with 5e12 /m2 density
        self._isNucleationSetup = False
        self.GBareaN0 = None
        self.GBedgeN0 = None
        self.GBcornerN0 = None
        self.dislocationN0 = None
        self.bulkN0 = None
        
        #Unit cell parameters
        self.aAlpha = None
        self.VaAlpha = None
        self.VmAlpha = None
        self.atomsPerCellAlpha = None
        self.atomsPerCellBeta = np.empty(len(self.phases), dtype=np.float32)
        self.VaBeta = np.empty(len(self.phases), dtype=np.float32)
        self.VmBeta = np.empty(len(self.phases), dtype=np.float32)
        self.Rmin = np.empty(len(self.phases), dtype=np.float32)
        
        #Free energy parameters
        self.gamma = np.empty(len(self.phases), dtype=np.float32)
        self.dG = [None for i in self.phases]
        self.interfacialComposition = [None for i in self.phases]

        #Beta function for nucleation rate
        if self.numberOfElements == 1:
            self._Beta = self._BetaBinary1
        else:
            self._Beta = self._BetaMulti
            self._betaFuncs = [None for p in phases]
            self._defaultBeta = 20

        #Stopping conditions
        self.clearStoppingConditions()

        #Coupling models
        self.clearCouplingModels()

    def phaseIndex(self, phase = None):
        '''
        Returns index of phase in list

        Parameters
        ----------
        phase : str (optional)
            Precipitate phase (defaults to None, which will return 0)
        '''
        return 0 if phase is None else np.where(self.phases == phase)[0][0]
            
    def reset(self):
        '''
        Resets simulation results
        This does not reset the model parameters, however, it will clear any stopping conditions
        '''
        self._resetArrays()
        self.xComp[0] = self.xInit
        self.dTemp = 0

        self._isSetup = False
        self._currY = None

        #Reset temperature array
        if np.isscalar(self.Tparameters):
            self.setTemperature(self.Tparameters)
        elif len(self.Tparameters) == 2:
            self.setTemperatureArray(*self.Tparameters)
        elif self.Tparameters is not None:
            self.setNonIsothermalTemperature(self.Tparameters)

        #Reset stopping conditions
        for sc in self._stoppingConditions:
            sc.reset()

    def _resetArrays(self):
        '''
        Resets and initializes arrays for all variables
            time, temperature
            matrix composition, equilibrium composition (alpha and beta)
            driving force, impingement factor, nucleation barrier, critical radius, nucleation radius
            nucleation rate, precipitate density
            average radius, average aspect ratio, volume fraction

        Extra variables include incubation offset and incubation sum

        Time dependent variables will be set up as either
            (iterations)                     time, temperature
            (iterations, elements)           composition
            (iterations, phases, elements)   eq composition, total precipitate composition
            (iterations, phases)             Everything else
            This is intended for appending arrays to always be on the first axis
        '''
        self.n = 0

        #Time
        self.time = np.zeros(1)

        #Temperature
        self.temperature = np.zeros(1)

        #Composition
        self.xComp = np.zeros((1, self.numberOfElements))                       #Matrix composition
        self.xEqAlpha = np.zeros((1, len(self.phases), self.numberOfElements))  #Equilibrium matrix composition
        self.xEqBeta = np.zeros((1, len(self.phases), self.numberOfElements))   #Equilibrium beta compostion

        #Nucleation
        self.dGs = np.zeros((1, len(self.phases)))                  #Driving force
        self.betas = np.zeros((1, len(self.phases)))                #Impingement rates (used for non-isothermal)
        self.incubationOffset = np.zeros(len(self.phases))          #Offset for incubation time (for non-isothermal precipitation)
        self.Gcrit = np.zeros((1, len(self.phases)))                #Height of nucleation barrier
        self.Rcrit = np.zeros((1, len(self.phases)))                #Critical radius
        self.Rad = np.zeros((1, len(self.phases)))                  #Radius of particles formed at each time step
        
        self.nucRate = np.zeros((1, len(self.phases)))              #Nucleation rate
        self.precipitateDensity = np.zeros((1, len(self.phases)))   #Number of nucleates
        
        #Average radius and precipitate fraction
        self.avgR = np.zeros((1, len(self.phases)))                 #Average radius
        self.avgAR = np.zeros((1, len(self.phases)))                #Mean aspect ratio
        self.betaFrac = np.zeros((1, len(self.phases)))             #Fraction of precipitate

        #Fconc - auxiliary array to store total solute composition of precipitates
        self.fConc = np.zeros((1, len(self.phases), self.numberOfElements))

        self._setEnum()
        self._packArrays()

        #Temporary storage variables
        self._precBetaTemp = [None for _ in range(len(self.phases))]    #Composition of nucleate (found from driving force)

    def _setEnum(self):
        '''
        Pseudo-enumeration

        This is just to keep a consistent list of IDs for each variable
        so we can grab the current values from varList
        '''
        self.TIME = 0
        self.TEMPERATURE = 1
        self.COMPOSITION = 2
        self.EQ_COMP_ALPHA = 3
        self.EQ_COMP_BETA = 4
        self.DRIVING_FORCE = 5
        self.IMPINGEMENT = 6
        self.G_CRIT = 7
        self.R_CRIT = 8
        self.R_NUC = 9
        self.NUC_RATE = 10
        self.PREC_DENS = 11
        self.R_AVG = 12
        self.AR_AVG = 13
        self.VOL_FRAC = 14
        self.FCONC = 15
        self.NUM_TERMS = 16

    def _packArrays(self):
        '''
        Create internal list of variables to solve for
        The "enumerators" in getEnum will serve as the indexing for this list
            Make sure the arrays in here and getEnum correspond to the same values
        '''
        self.varList = [
                self.time, 
                self.temperature, 
                self.xComp, 
                self.xEqAlpha, 
                self.xEqBeta,
                self.dGs, 
                self.betas, 
                self.Gcrit, 
                self.Rcrit, 
                self.Rad,
                self.nucRate, 
                self.precipitateDensity,
                self.avgR, 
                self.avgAR, 
                self.betaFrac,
                self.fConc
                ]

    def _getVarDict(self):
        '''
        Returns mapping of { variable name : attribute name } for saving
        The variable name will be the name in the .npz file
        '''
        saveDict = {
            'elements': 'elements',
            'phases': 'phases',
            'time': 'time',
            'temperature': 'temperature',
            'composition': 'xComp',
            'xEqAlpha': 'xEqAlpha',
            'xEqBeta': 'xEqBeta',
            'drivingForce': 'dGs',
            'impingement': 'betas',
            'Gcrit': 'Gcrit',
            'Rcrit': 'Rcrit',
            'nucRadius': 'Rad',
            'nucRate': 'nucRate',
            'precipitateDensity': 'precipitateDensity',
            'avgRadius': 'avgR',
            'avgAspectRatio': 'avgAR',
            'volFrac': 'betaFrac',
            'fConc': 'fConc',
        }
        return saveDict
    
    def load(filename):
        '''
        Loads data from filename and returns a PrecipitateModel
        '''
        data = np.load(filename)
        model = PrecipitateBase(data['phases'], data['elements'])
        model._loadData(data)
        return model
    
    def _appendArrays(self, newVals):
        '''
        Appends new values to the variable list
        NOTE: newVals must correspond to the same order as _packArrays with first axis as 1
            Ex rCrit is (n, phases) so corresponding new value should be (1, phases)
        Since np append creates a new variable in memory, we have to reassign each term, then pack them into varList again
            TODO: it would be nice to reduce the number of times it copies, perhaps by preallocating some amount (say 1000)
                    for each array and if we have not reached the end of the array, just stick the values at the latest index
                    but once we reach the end of the array, we would append another 1000
                    The after solving, we could clean up the arrays, or just use self.n to state where the end of the simulation is
        I suppose we could make a list of str for each variable and call setattr
        '''
        self.time = np.append(self.time, newVals[self.TIME], axis=0)
        self.temperature = np.append(self.temperature, newVals[self.TEMPERATURE], axis=0)
        self.xComp = np.append(self.xComp, newVals[self.COMPOSITION], axis=0)
        self.xEqAlpha = np.append(self.xEqAlpha, newVals[self.EQ_COMP_ALPHA], axis=0)
        self.xEqBeta = np.append(self.xEqBeta, newVals[self.EQ_COMP_BETA], axis=0)
        self.dGs = np.append(self.dGs, newVals[self.DRIVING_FORCE], axis=0)
        self.betas = np.append(self.betas, newVals[self.IMPINGEMENT], axis=0)
        self.Gcrit = np.append(self.Gcrit, newVals[self.G_CRIT], axis=0)
        self.Rcrit = np.append(self.Rcrit, newVals[self.R_CRIT], axis=0)
        self.Rad = np.append(self.Rad, newVals[self.R_NUC], axis=0)
        self.nucRate = np.append(self.nucRate, newVals[self.NUC_RATE], axis=0)
        self.precipitateDensity = np.append(self.precipitateDensity, newVals[self.PREC_DENS], axis=0)
        self.avgR = np.append(self.avgR, newVals[self.R_AVG], axis=0)
        self.avgAR = np.append(self.avgAR, newVals[self.AR_AVG], axis=0)
        self.betaFrac = np.append(self.betaFrac, newVals[self.VOL_FRAC], axis=0)
        self.fConc = np.append(self.fConc, newVals[self.FCONC], axis=0)
        self._packArrays()
        self.n += 1

    def resetConstraints(self):
        '''
        Default values for contraints
        '''
        self.minRadius = 3e-10
        self.maxTempChange = 1

        self.maxDTFraction = 1e-2
        self.minDTFraction = 1e-5

        #Constraints on maximum time step
        self.checkTemperature = True
        self.maxNonIsothermalDT = 1

        self.checkPSD = True
        self.maxDissolution = 1e-3

        self.checkRcrit = True
        self.maxRcritChange = 0.01

        self.checkNucleation = True
        self.maxNucleationRateChange = 0.5
        self.minNucleationRate = 1e-5

        self.checkVolumePre = True
        self.maxVolumeChange = 0.001
        
        self.minComposition = 0

        self.minNucleateDensity = 1e-10

        #TODO: may want to test more to see if this value should be lower or higher
        #This will attempt to increase the time by 0.1%
        #This also only affects the sim if the calculated dt is extremely large
        #So probably only when nucleation rate is 0 will this matter
        #This roughly corresponds to 1e4 steps over 5-7 orders of magnitude on a log time scale
        self.dtScale = 1e-3

    def setConstraints(self, **kwargs):
        '''
        Sets constraints

        Possible constraints:
        ---------------------
        minRadius - minimum radius to be considered a precipitate (1e-10 m)
        maxTempChange - maximum temperature change before lookup table is updated (only for Euler in binary case) (1 K)

        maxDTFraction - maximum time increment allowed as a fraction of total simulation time (0.1)
        minDTFraction - minimum time increment allowed as a fraction of total simulation time (1e-5)

        checkTemperature - checks max temperature change (True)
        maxNonIsothermalDT - maximum time step when temperature is changing (1 second)

        checkPSD - checks maximum growth rate for particle size distribution (True)
        maxDissolution - maximum relative volume fraction of precipitates allowed to dissolve in a single time step (0.01)

        checkRcrit - checks maximum change in critical radius (False)
        maxRcritChange - maximum change in critical radius (as a fraction) per single time step (0.01)

        checkNucleation - checks maximum change in nucleation rate (True)
        maxNucleationRateChange - maximum change in nucleation rate (on log scale) per single time step (0.5)
        minNucleationRate - minimum nucleation rate to be considered for checking time intervals (1e-5)

        checkVolumePre - estimates maximum volume change (True)
        checkVolumePost - checks maximum calculated volume change (True)
        maxVolumeChange - maximum absolute value that volume fraction can change per single time step (0.001)

        minNucleateDensity - minimum nucleate density to consider nucleation to have occurred (1e-5)
        dtScale - scaling factor to attempt to progressively increase dt over time
        '''
        for key, value in kwargs.items():
            setattr(self, key, value)

    def setBetaBinary(self, functionType = 1):
        '''
        Sets function for beta calculation in binary systems
            1 for implementation seen in Perez et al, 2008 (default)
            2 for implementation similar to multicomponent systems

        If using a multicomponent system, the beta function defaults to the 2nd
            So this function will not do anything

        Parameters
        ----------
        functionType : int
            ID for function
                1 for implementation seen in Perez et al, 2008 (default)
                2 for implementation similar to multicomponent systems
        '''
        if self.numberOfElements == 1:
            if functionType == 2:
                self.beta = self._BetaBinary2
            else:
                self.beta = self._BetaBinary1

    def setInitialComposition(self, xInit):
        '''
        Parameters
        
        xInit : float or array
            Initial composition of parent matrix phase in atomic fraction
            Use float for binary system and array of solutes for multicomponent systems
        '''
        self.xInit = xInit
        self.xComp[0] = xInit
        
    def setInterfacialEnergy(self, gamma, phase = None):
        '''
        Parameters
        ----------
        gamma : float
            Interfacial energy between precipitate and matrix in J/m2
        phase : str (optional)
            Phase to input interfacial energy (defaults to first precipitate in list)
        '''
        index = self.phaseIndex(phase)
        self.gamma[index] = gamma
        
    def resetAspectRatio(self, phase = None):
        '''
        Resets aspect ratio variables of defined phase to default

        phase : str (optional)
            Phase to consider (defaults to first precipitate in list)
        '''
        index = self.phaseIndex(phase)
        self.shapeFactors[index].setSpherical()

    def setPrecipitateShape(self, precipitateShape, phase = None, ratio = 1):
        '''
        Sets precipitate shape to user-defined shape

        Parameters
        ----------
        precipitateShape : int
            Precipitate shape (ShapeFactor.SPHERE, NEEDLE, PLATE or CUBIC)
        phase : str (optional)
            Phase to consider (defaults to first precipitate in list)
        ratio : float (optional)
            Aspect ratio of precipitate (long axis / short axis)
            If float, must be greater than 1
            If function, must take in radius as input and output float greater than 1
        '''
        index = self.phaseIndex(phase)
        self.shapeFactors[index].setPrecipitateShape(precipitateShape, ratio)

    def _setVolume(self, value, valueType: VolumeParameter, atomsPerCell):
        '''
        Private function that returns Vm, Va, a, atomsPerCell given a VolumeParameter and atomsPerCell

        Parameters
        ----------
        value : float
            Value for volume parameters (lattice parameter, atomic (unit cell) volume or molar volume)
        valueType : VolumeParameter
            States what volume term that value is
        atomsPerCell : int
            Number of atoms in the unit cell
        '''
        if valueType == VolumeParameter.MOLAR_VOLUME:
            Vm = value
            Va = atomsPerCell * Vm / self.avo
            a = np.cbrt(Va)
        elif valueType == VolumeParameter.ATOMIC_VOLUME:
            Va = value
            Vm = Va * self.avo / atomsPerCell
            a = np.cbrt(Va)
        elif valueType == VolumeParameter.LATTICE_PARAMETER:
            a = value
            Va = a**3
            Vm = Va * self.avo / atomsPerCell
        return Vm, Va, a, atomsPerCell
    
    def setVolumeAlpha(self, value, valueType: VolumeParameter, atomsPerCell):
        '''
        Sets volume parameters for parent phase

        Parameters
        ----------
        value : float
            Value for volume parameters (lattice parameter, atomic (unit cell) volume or molar volume)
        valueType : VolumeParameter
            States what volume term that value is
        atomsPerCell : int
            Number of atoms in the unit cell
        '''
        self.VmAlpha, self.VaAlpha, self.aAlpha, self.atomsPerCellAlpha = self._setVolume(value, valueType, atomsPerCell)

    def setVolumeBeta(self, value, valueType: VolumeParameter, atomsPerCell, phase = None):
        '''
        Sets volume parameters for precipitate phase

        Parameters
        ----------
        value : float
            Value for volume parameters (lattice parameter, atomic (unit cell) volume or molar volume)
        valueType : VolumeParameter
            States what volume term that value is
        atomsPerCell : int
            Number of atoms in the unit cell
        phase : str (optional)
            Phase to consider (defaults to first precipitate in list)
        '''
        index = self.phaseIndex(phase)
        self.VmBeta[index], self.VaBeta[index], _, self.atomsPerCellBeta[index] = self._setVolume(value, valueType, atomsPerCell)

    def setNucleationDensity(self, grainSize = 100, aspectRatio = 1, dislocationDensity = 5e12, bulkN0 = None):
        '''
        Sets grain size and dislocation density which determines the available nucleation sites
        
        Parameters
        ----------
        grainSize : float (optional)
            Average grain size in microns (default at 100um if this function is not called)
        aspectRatio : float (optional)
            Aspect ratio of grains (default at 1)
        dislocationDensity : float (optional)
            Dislocation density (m/m3) (default at 5e12)
        bulkN0 : float (optional)
            This allows for the use to override the nucleation site density for bulk precipitation
            By default (None), this is calculated by the number of lattice sites containing a solute atom
            However, for calibration purposes, it may be better to set the nucleation site density manually
        '''
        self.grainSize = grainSize * 1e-6
        self.grainAspectRatio = aspectRatio
        self.dislocationDensity = dislocationDensity

        self.bulkN0 = bulkN0
        self._isNucleationSetup = True

    def _getNucleationDensity(self):
        '''
        Calculates nucleation density
        This is separated from setting nucleation density to
            allow it to be called right before the simulation starts
        '''
        #Set bulk nucleation site to the number of solutes per unit volume
        #   This is the represent that any solute atom can be a nucleation site
        #NOTE: some texts will state the bulk nucleation sites to just be the number
        #       of lattice sites per unit volume. The justification for this would be 
        #       the solutes can diffuse around to any lattice site and nucleate there
        if self.bulkN0 is None:
            if self.numberOfElements == 1:
                self.bulkN0 = self.xComp[0] * (self.avo / self.VmAlpha)
            else:
                self.bulkN0 = np.amin(self.xComp[0,:]) * (self.avo / self.VmAlpha)

        self.dislocationN0 = self.dislocationDensity * (self.avo / self.VmAlpha)**(1/3)
        
        if self.grainSize != np.inf:
            #Number of lattice sites on grain boundaries (#/m3)
            if self.GBareaN0 is None:
                self.GBareaN0 = (6 * np.sqrt(1 + 2 * self.grainAspectRatio**2) + 1 + 2 * self.grainAspectRatio) / (4 * self.grainAspectRatio * self.grainSize)
                self.GBareaN0 *= (self.avo / self.VmAlpha)**(2/3)
            #Number of lattice sites on grain edges (#/m3)
            if self.GBedgeN0 is None:
                self.GBedgeN0 = 2 * (np.sqrt(2) + 2 * np.sqrt(1 + self.grainAspectRatio**2)) / (self.grainAspectRatio * self.grainSize**2)
                self.GBedgeN0 *= (self.avo / self.VmAlpha)**(1/3)
            #Number of lattice sites on grain corners (which is just the number of corners) (#/m3)
            if self.GBcornerN0 is None:
                self.GBcornerN0 = 12 / (self.grainAspectRatio * self.grainSize**3)
        else:
            self.GBareaN0 = 0
            self.GBedgeN0 = 0
            self.GBcornerN0 = 0
        
    def setNucleationSite(self, site, phase = None):
        '''
        Sets nucleation site type for specified phase
        If site type is grain boundaries, edges or corners, the phase morphology will be set to spherical and precipitate shape will depend on wetting angle
        
        Parameters
        ----------
        site : str
            Type of nucleation site
            Options are 'bulk', 'dislocations', 'grain_boundaries', 'grain_edges' and 'grain_corners'
        phase : str (optional)
            Phase to consider (defaults to first precipitate in list)
        '''
        index = self.phaseIndex(phase)

        self.GB[index].setNucleationType(site)
        
        if self.GB[index].nucleationSiteType != GBFactors.BULK and self.GB[index].nucleationSiteType != GBFactors.DISLOCATION:
            self.shapeFactors[index].setSpherical()
            
    def _setGBfactors(self):
        '''
        Calcualtes factors for bulk or grain boundary nucleation
        This is separated from setting the nucleation sites to allow
        it to be called right before simulation
        '''
        for p in range(len(self.phases)):
            self.GB[p].setFactors(self.GBenergy, self.gamma[p])
                    
    def _GBareaRemoval(self, p):
        '''
        Returns factor to multiply radius by to give the equivalent radius of circles representing the area of grain boundary removal

        Parameters
        ----------
        p : int
            Index for phase
        '''
        if self.GB[p].nucleationSiteType == GBFactors.BULK or self.GB[p].nucleationSiteType == GBFactors.DISLOCATION:
            return 1
        else:
            return np.sqrt(self.GB[p].gbRemoval / np.pi)
            
    def setParentPhases(self, phase, parentPhases):
        '''
        Sets parent precipitates at which a precipitate can nucleate on the surface of
        
        Parameters
        ----------
        phase : str
            Precipitate phase of interest that will nucleate
        parentPhases : list
            Phases that the precipitate of interest can nucleate on the surface of
        '''
        index = self.phaseIndex(phase)
        for p in parentPhases:
            self.parentPhases[index].append(self.phaseIndex(p))
           
    def setGrainBoundaryEnergy(self, energy):
        '''
        Grain boundary energy - this will decrease the critical radius as some grain boundaries will be removed upon nucleation

        Parameters
        ----------
        energy : float
            GB energy in J/m2

        Default upon initialization is 0.3
        Note: GBenergy of 0 is equivalent to bulk precipitation
        '''
        self.GBenergy = energy
        
    def setTheta(self, theta, phase = None):
        '''
        This is a scaling factor for the incubation time calculation, default is 2

        Incubation time is defined as 1 / \theta * \beta * Z^2
        \theta differs by derivation. By default, this is set to 2 following the
        Feder derivation. In the Wakeshima derivation, \theta is 4pi

        Parameters
        ----------
        theta : float
        phase : str (optional)
            Phase to consider (defaults to first precipitate in list)
        '''
        index = self.phaseIndex(phase)
        self.theta[index] = theta

    def setTemperature(self, temperature):
        '''
        Sets temperature parameter

        Options:
            temperature : float
                Isothermal temperature
            temperature : function
                Function takes in time in seconds and returns temperature
            temperature : [times, temps]
                Temperature will be interpolated between the times and temps list
                Each index in the lists will correspond to the time that temperature is reached
                Ex. [0, 15, 20], [100, 500, 400]
                    Temperature starts at 100 and ramps to 500, reaching it at 15 hours
                    Then temperature will drop to 400, reaching it at 20 hours
        '''
        self.Tparameters = temperature
        self.temperature[0] = self.getTemperature(0)
        if np.isscalar(temperature):
            self._incubation = self._incubationIsothermal
        else:
            self._incubation = self._incubationNonIsothermal

    def getTemperature(self, t):
        '''
        Gets temperature at time t

        Options:
            Options:
            temperature : float
                Returns temperature
            temperature : function
                Returns evaluated temperature function at time t
            temperature : [times, temps]
                If t < time[0] -> return first temperature
                If t > time[-1] -> return last temperature
                Else, find the two times that t is between and interpolate
        '''
        if np.isscalar(self.Tparameters):
            return self.Tparameters
        elif len(self.Tparameters) == 2:
            if t/3600 < self.Tparameters[0][0]:
                return self.Tparameters[1][0]
            for i in range(len(self.Tparameters[0])-1):
                if t/3600 >= self.Tparameters[0][i] and t/3600 < self.Tparameters[0][i+1]:
                    t0, tf, T0, Tf = self.Tparameters[0][i], self.Tparameters[0][i+1], self.Tparameters[1][i], self.Tparameters[1][i+1]
                    return (Tf - T0) / (tf - t0) * (t/3600 - t0) + T0
            return self.Tparameters[1][-1]
        elif self.Tparameters is not None:
            return self.Tparameters(t)
        else:
            return None
        
    def setStrainEnergy(self, strainEnergy, phase = None, calculateAspectRatio = False):
        '''
        Sets strain energy class to precipitate

        Parameters
        ----------
        strainEnergy : StrainEnergy object
        phase : str
            Precipitate phase of interest that will nucleate
        calculateAspectRatio : bool
            Will use strain energy to get aspect ratio if True
        '''
        index = self.phaseIndex(phase)
        self.strainEnergy[index] = strainEnergy
        self.calculateAspectRatio[index] = calculateAspectRatio

    def _setupStrainEnergyFactors(self):
        ''''
        For each phase, the strain energy calculation will be set to assume
        a spherical, cubic or ellipsoidal shape depending on the defined shape factors
        '''
        for i in range(len(self.phases)):
            self.strainEnergy[i].setup()
            if self.strainEnergy[i].type != StrainEnergy.CONSTANT:
                if self.shapeFactors[i].particleType == ShapeFactor.SPHERE:
                    self.strainEnergy[i].setSpherical()
                elif self.shapeFactors[i].particleType == ShapeFactor.CUBIC:
                    self.strainEnergy[i].setCuboidal()
                else:
                    self.strainEnergy[i].setEllipsoidal()

    def setDiffusivity(self, diffusivity):
        '''
        For binary systems only

        Parameters
        ----------
        diffusivity : function taking 
            Composition and temperature (K) and returning diffusivity (m2/s)
            Function must have inputs in this order: f(x, T)
        '''
        self.Diffusivity = diffusivity

    def setInfinitePrecipitateDiffusivity(self, infinite, phase = None):
        '''
        Sets whether to assuming infinitely fast or no diffusion in phase

        Parameters
        ----------
        infinite : bool
            True will assume infinitely fast diffusion
            False will assume no diffusion
        phase : str (optional)
            Phase to consider (defaults to first precipitate in list)
            Use 'all' to apply to all phases
        '''
        if phase == 'all':
            self.infinitePrecipitateDiffusion = [infinite for i in range(len(self.phases))]
        else:
            index = self.phaseIndex(phase)
            self.infinitePrecipitateDiffusion[index] = infinite

    def setThermodynamics(self, therm, phase = None, removeCache = False, addDiffusivity = True):
        '''
        Parameters
        ----------
        therm : Thermodynamics class
        phase : str (optional)
            Phase to consider (defaults to first precipitate in list)
        removeCache : bool (optional)
            Will not cache equilibrium results if True (defaults to False)
        addDiffusivity : bool (optional)
            For binary systems, will add diffusivity functions from the database if present
            Defaults to True
        '''
        index = self.phaseIndex(phase)
        self.dG[index] = lambda x, T, removeCache = removeCache: therm.getDrivingForce(x, T, precPhase=phase, training = removeCache, returnComp = True)
        
        if self.numberOfElements == 1:
            self.interfacialComposition[index] = lambda x, T: therm.getInterfacialComposition(x, T, precPhase=phase)
            if (therm.mobCallables is not None or therm.diffCallables is not None) and addDiffusivity:
                self.Diffusivity = lambda x, T, removeCache = removeCache: therm.getInterdiffusivity(x, T, removeCache = removeCache)
        else:
            self.interfacialComposition[index] = lambda x, T, dG, R, gExtra, removeCache = removeCache, searchDir = None: therm.getGrowthAndInterfacialComposition(x, T, dG, R, gExtra, precPhase=phase, training = False, searchDir = searchDir)
            self._betaFuncs[index] = lambda x, T, removeCache = removeCache: therm.impingementFactor(x, T, precPhase=phase, training = False)

    def setSurrogate(self, surr, phase = None):
        '''
        Parameters
        ----------
        surr : Surrogate class
        phase : str (optional)
            Phase to consider (defaults to first precipitate in list)
        '''
        index = self.phaseIndex(phase)
        self.dG[index] = surr.getDrivingForce
        
        if self.numberOfElements == 1:
            self.interfacialComposition[index] = surr.getInterfacialComposition
        else:
            self.interfacialComposition[index] = surr.getGrowthAndInterfacialComposition
            self._betaFuncs[index] = surr.impingementFactor

    def particleGibbs(self, radius, phase = None):
        '''
        Returns Gibbs Thomson contribution of a particle given its radius

        Parameters
        ----------
        radius : float or array
            Precipitate radius
        phase : str (optional)
            Phase to consider (defaults to first precipitate in list)
        '''
        index = self.phaseIndex(phase)
        return self.VmBeta[index] * (self.strainEnergy[index].strainEnergy(self.shapeFactors[index].normalRadii(radius)) + 2 * self.shapeFactors[index].thermoFactor(radius) * self.gamma[index] / radius)

    def neglectEffectiveDiffusionDistance(self, neglect = True):
        '''
        Whether or not to account for effective diffusion distance dependency on the supersaturation
        By default, effective diffusion distance is considered
        
        Parameters
        ----------
        neglect : bool (optional)
            If True (default), will assume effective diffusion distance is particle radius
            If False, will calculate correction factor from Chen, Jeppson and Agren (2008)
        '''
        self.effDiffDistance = self.effDiffFuncs.noDiffusionDistance if neglect else self.effDiffFuncs.effectiveDiffusionDistance

    def addStoppingCondition(self, condition, mode = 'or'):
        '''
        Adds condition to stop simulation when condition is met

        Parameters
        ----------
        condition: PrecipitateStoppingCondition
        mode: str
            'or' or 'and
            Conditions with 'or' will stop the simulation when at least one condition is met
            Conditions with 'and' will stop the simulation when all conditions are met
        '''
        self._stoppingConditions.append(condition)
        if mode == 'or':
            self._stopConditionMode.append(True)
        else:
            self._stopConditionMode.append(False)
        
    def clearStoppingConditions(self):
        '''
        Clears all stopping conditions
        '''
        self._stoppingConditions = []
        self._stopConditionMode = []

    def setup(self):
        '''
        Sets up hidden parameters before solving
            Nucleation site density
            Grain boundary factors
            Strain energy
        '''
        if self._isSetup:
            return
        
        if not self._isNucleationSetup:
            #Set nucleation density assuming grain size of 100 um and dislocation density of 5e12 m/m3 (Thermocalc default)
            print('Nucleation density not set.\nSetting nucleation density assuming grain size of {:.0f} um and dislocation density of {:.0e} #/m2'.format(100, 5e12))
            self.setNucleationDensity(100, 1, 5e12)
        for p in range(len(self.phases)):
            self.Rmin[p] = self.minRadius
        self._getNucleationDensity()
        self._setGBfactors()
        self._setupStrainEnergyFactors()
        self._isSetup = True

    def printHeader(self):
        '''
        Overloads printHeader from GenericModel to do nothing
        since status displays the necessary outputs
        '''
        return

    def printStatus(self, iteration, modelTime, simTimeElapsed):
        '''
        Prints various terms at latest step

        Will print:
            Model time, simulation time, temperature, matrix composition
            For each phase
                Phase name, precipitate density, volume fraction, avg radius and driving force
        '''
        i = len(self.time)-1
        #For single element, we just print the composition as matrix comp in terms of the solute
        if self.numberOfElements == 1:
            print('N\tTime (s)\tSim Time (s)\tTemperature (K)\tMatrix Comp')
            print('{:.0f}\t{:.1e}\t\t{:.1f}\t\t{:.0f}\t\t{:.4f}\n'.format(i, modelTime, simTimeElapsed, self.temperature[i], 100*self.xComp[i,0]))
        #For multicomponent systems, print each element
        else:
            compStr = 'N\tTime (s)\tSim Time (s)\tTemperature (K)\t'
            compValStr = '{:.0f}\t{:.1e}\t\t{:.1f}\t\t{:.0f}\t\t'.format(i, modelTime, simTimeElapsed, self.temperature[i])
            for a in range(self.numberOfElements):
                compStr += self.elements[a] + '\t'
                compValStr += '{:.4f}\t'.format(100*self.xComp[i,a])
            compValStr += '\n'
            print(compStr)
            print(compValStr)

        #Print status of each phase
        print('\tPhase\tPrec Density (#/m3)\tVolume Frac\tAvg Radius (m)\tDriving Force (J/mol)')
        for p in range(len(self.phases)):
            print('\t{}\t{:.3e}\t\t{:.4f}\t\t{:.4e}\t{:.4e}'.format(self.phases[p], self.precipitateDensity[i,p], 100*self.betaFrac[i,p], self.avgR[i,p], self.dGs[i,p]*self.VmBeta[p]))
        print('')

    def preProcess(self):
        '''
        Store array for non-derivative terms (which is everything except for the PBM models)

        We use these terms for the first step of the iterators (for Euler, this is all the steps)
            For RK4, these terms will be recalculated in dXdt
        '''
        self._currY = None
        return
    
    def _calculateDependentTerms(self, t, x):
        '''
        Gets all dependent terms (everything but PBM variables) that are needed to find dXdt

        Steps:
            1. Mass balance
            2. Driving force - must be done after mass balance to get the current matrix composition
            3. Growth rate - must be done after driving force since dG is needed in multicomponent systems
            4. Nucleation rate
            5. Nucleate radius - must be done after nucleation rate since derived classes can change nucleation rate

        For the first iteration, self._currY will be None from the preProcess function, in this case, we want
            to just grab the latest values to avoid double calculations
        '''
        self._processX(x)
        if self._currY is None:
            #print('start iteration')
            self._currY = [np.array([self.varList[i][self.n]]) for i in range(self.NUM_TERMS)]
        else:
            self._currY[self.TIME] = np.array([t])
            self._currY[self.TEMPERATURE] = np.array([self.getTemperature(t)])
            self._calcMassBalance(t, x)
            self._calcDrivingForce(t, x)
            self._growthRate()
            self._calcNucleationRate(t, x)
            self._setNucleateRadius(t)      #Must be done afterwards since derived classes can change nucRate
        #print(self._currY)

    def getdXdt(self, t, x):
        '''
        Gets dXdt as a list for each phase

        For the eulerian implementation, this is dn_i/dt for the bins in PBM for each phase
        '''
        self._calculateDependentTerms(t, x)
        dXdt = self._getdXdt(t, x)
        return dXdt

    def postProcess(self, t, x):
        '''
        1) Updates internal arrays with new values of t and x
        2) Updates particle size distribution
        3) Updates coupled models
        4) Check stopping conditions
        5) Return new values and whether to stop the model
        '''
        self._calculateDependentTerms(t, x)
        self._appendArrays(self._currY)

        #Update particle size distribution (this includes adding bins, resizing bins, etc)
        #Should be agnostic of eulerian or lagrangian implementations
        self._updateParticleSizeDistribution(t, x)

        #Update coupled models
        self.updateCoupledModels()

        #Check stopping conditions
        orCondition = False
        andCondition = True
        numAndCondition = 0
        for i in range(len(self._stoppingConditions)):
            self._stoppingConditions[i].testCondition(self)
            if self._stopConditionMode[i]:
                orCondition = orCondition or self._stoppingConditions[i].isSatisfied()
            else:
                andCondition = andCondition and self._stoppingConditions[i].isSatisfied()
                numAndCondition += 1

        #If no and conditions, then andCondition will still be True, so set to False
        if numAndCondition == 0:
            andCondition = False

        stop = orCondition or andCondition

        return self.getCurrentX()[1], stop
    
    def _processX(self, x):
        return NotImplementedError()
    
    def _calcMassBalance(self, t, x):
        return NotImplementedError()
    
    def _getdXdt(self, t, x):
        return NotImplementedError()
    
    def _updateParticleSizeDistribution(self, t, x):
        return NotImplementedError()
    
    def _calcDrivingForce(self, t, x):
        '''
        Driving force is defined in terms of J/m3

        Calculation is dG_ch / V_m - dG_el
            dG_ch - chemical driving force
            V_m - molar volume
            dG_el - elastic strain energy (always reduces driving force)
                I guess there could be a case where it increases the driving force if
                the matrix is prestrained and the precipitate releases stress, but this should
                be handled in the ElasticFactors module

        If driving force is positive (precipitation is favorable)
            Calculate Rcrit and Gcrit based off the nucleation site type

        This will also calculate critical radius (Rcrit) and nucleation barrier (Gcrit)
        '''
        #Get variables
        dGs = np.zeros((1,len(self.phases)))
        Rcrit = np.zeros((1,len(self.phases)))
        Gcrit = np.zeros((1,len(self.phases)))
        if self.numberOfElements == 1:
            xComp = self._currY[self.COMPOSITION][0,0]
        else:
            xComp = self._currY[self.COMPOSITION][0]
        T = self._currY[self.TEMPERATURE][0]

        for p in range(len(self.phases)):
            dGs[0,p], self._precBetaTemp[p] = self.dG[p](xComp, T)
            dGs[0,p] /= self.VmBeta[p]
            dGs[0,p] -= self.strainEnergy[p].strainEnergy(self.shapeFactors[p].normalRadii(self.Rcrit[self.n, p]))
            if self.betaFrac[self.n, p] < 1 and dGs[0,p] >= 0:
                #Calculate critical radius
                #For bulk or dislocation nucleation sites, use previous critical radius to get aspect ratio
                if self.GB[p].nucleationSiteType == GBFactors.BULK or self.GB[p].nucleationSiteType == GBFactors.DISLOCATION:
                    Rcrit[0,p] = np.amax((2 * self.shapeFactors[p].thermoFactor(self.Rcrit[self.n, p]) * self.gamma[p] / dGs[0,p], self.Rmin[p]))
                    Gcrit[0,p] = (4 * np.pi / 3) * self.gamma[p] * Rcrit[0,p]**2

                #If nucleation is on a grain boundary, then use the critical radius as defined by the grain boundary type    
                else:
                    Rcrit[0,p] = np.amax((self.GB[p].Rcrit(dGs[0,p]), self.Rmin[p]))
                    Gcrit[0,p] = self.GB[p].Gcrit(dGs[0,p], Rcrit[0,p])

        self._currY[self.DRIVING_FORCE] = dGs
        self._currY[self.R_CRIT] = Rcrit
        self._currY[self.G_CRIT] = Gcrit

    def _calcNucleationRate(self, t, x):
        '''
        nucleation rate is defined as dn_nuc/dt = N_0 Z beta exp(-G/kBt) * exp(-tau/t)
        '''
        gCrit = self._currY[self.G_CRIT][0]
        T = self._currY[self.TEMPERATURE][0]
        dg = self._currY[self.DRIVING_FORCE][0]

        betas = np.zeros((1,len(self.phases)))
        nucRate = np.zeros((1,len(self.phases)))
        for p in range(len(self.phases)):
            #If driving force is negative, then nucleation rate is 0
            if dg[p] < 0:
                continue

            Z = self._Zeldovich(p)
            betas[0,p] = self._Beta(p)

            #If beta is 0, then nucRate is 0 and no need to do anymore calculation
            if betas[0,p] == 0:
                continue
            
            #Incubation time, either isothermal or nonisothermal
            incubation = self._incubation(t, p, Z, betas[0])
            if incubation > 1:
                incubation = 1

            nucRate[0,p] = Z * betas[0,p] * np.exp(-gCrit[p] / (self.kB * T)) * incubation

        self._currY[self.IMPINGEMENT] = betas
        self._currY[self.NUC_RATE] = nucRate

    def _Zeldovich(self, p):
        '''
        Zeldovich factor - probability that cluster at height of nucleation barrier will continue to grow
        '''
        rCrit = self._currY[self.R_CRIT][0]
        T = self._currY[self.TEMPERATURE][0]
        if rCrit[p] == 0:
            return 0
        return np.sqrt(3 * self.GB[p].volumeFactor / (4 * np.pi)) * self.VmBeta[p] * np.sqrt(self.gamma[p] / (self.kB * T)) / (2 * np.pi * self.avo * rCrit[p]**2)
        
    def _BetaBinary1(self, p):
        '''
        Impingement rate for binary systems using Perez et al
        '''
        rCrit = self._currY[self.R_CRIT][0]
        xComp = self._currY[self.COMPOSITION][0][0]
        T = self._currY[self.TEMPERATURE][0]
        return self.GB[p].areaFactor * rCrit[p]**2 * self.xComp[0] * self.Diffusivity(xComp, T) / self.aAlpha**4

    def _BetaBinary2(self, p):
        '''
        Impingement rate for binary systems taken from Thermocalc prisma documentation
        This will follow the same equation as with _BetaMulti; however, some simplications can be made based off the summation contraint
        '''
        xComp = self._currY[self.COMPOSITION][0][0]
        xEqAlpha = self._currY[self.EQ_COMP_ALPHA][0]
        xEqBeta = self._currY[self.EQ_COMP_BETA][0]
        rCrit = self._currY[self.R_CRIT][0]
        T = self._currY[self.TEMPERATURE][0]
        D = self.Diffusivity(xComp, T)
        Dfactor = (xEqBeta[p] - xEqAlpha[p])**2 / (xEqAlpha[p]*D) + (xEqBeta[p] - xEqAlpha[p])**2 / ((1 - xEqAlpha[p])*D)
        return self.GB[p].areaFactor * rCrit[p]**2 * (1/Dfactor) / self.aAlpha**4
            
    def _BetaMulti(self, p):
        '''
        Impingement rate for multicomponent systems
        '''
        if self._betaFuncs[p] is None:
            return self._defaultBeta
        else:
            xComp = self._currY[self.COMPOSITION][0]
            T = self._currY[self.TEMPERATURE][0]
            beta = self._betaFuncs[p](xComp, T)
            if beta is None:
                return self.betas[p]
            else:
                rCrit = self._currY[self.R_CRIT][0]
                return (self.GB[p].areaFactor * rCrit[p]**2 / self.aAlpha**4) * beta

    def _incubationIsothermal(self, t, p, Z, betas):
        '''
        Incubation time for isothermal conditions
        '''
        tau = 1 / (self.theta[p] * (betas[p] * Z**2))
        return np.exp(-tau / t)
        
    def _incubationNonIsothermal(self, t, p, Z, betas):
        '''
        Incubation time for non-isothermal conditions
        This must match isothermal conditions if the temperature is constant

        Solve for tau by: integral(beta(t-t0)) from 0 to tau = 1/theta*Z(tau)^2

        Then it's exp(-tau/t) like the isothermal behavior
        '''
        T = self._currY[self.TEMPERATURE][0]
        startIndex = int(self.incubationOffset[p])
        LHS = 1 / (self.theta[p] * Z**2 * (T / self.temperature[startIndex:self.n+1]))

        RHS = np.cumsum(self.betas[startIndex+1:self.n+1,p] * (self.time[startIndex+1:self.n+1] - self.time[startIndex:self.n]))
        if len(RHS) == 0:
            RHS = self.betas[self.n,p] * (self.time[startIndex:] - self.time[startIndex])
        else:
            RHS = np.concatenate((RHS, [RHS[-1] + betas[p] * (t - self.time[startIndex])]))

        #Test for intersection
        diff = RHS - LHS
        signChange = np.sign(diff[:-1]) != np.sign(diff[1:])

        #If no intersection
        if not any(signChange):
            #If RHS > LHS, then intersection is at t = 0
            if diff[0] > 0:
                tau = 0
            #Else, RHS intersects LHS beyond simulation time
            #Extrapolate integral of RHS from last point to intersect LHS
            #integral(beta(t-t0)) from t0 to ti + beta_i * (tau - (ti - t0)) = 1 / theta * Z(tau+t0)^2
            else:
                tau = LHS[-1] / betas[p] - RHS[-1] / betas[p] + (t - self.time[startIndex])
        else:
            tau = self.time[startIndex:-1][signChange][0] - self.time[startIndex]

        return np.exp(-tau / (t - self.time[startIndex]))
    
    def _setNucleateRadius(self, t):
        '''
        Adds 1/2 * sqrt(kb T / pi gamma) to critical radius to ensure they grow when growth rates are calculated
        '''
        nucRate = self._currY[self.NUC_RATE][0]
        T = self._currY[self.TEMPERATURE][0]
        dt = t - self.time[self.n]
        Rcrit = self._currY[self.R_CRIT][0]
        Rad = np.zeros((1,len(self.phases)))
        for p in range(len(self.phases)):
            #If nucleates form, then calculate radius of precipitate
            #Radius is set slightly larger so precipitate
            dt = 0.01 if self.n == 0 else self.time[self.n] - self.time[self.n-1]
            if nucRate[p]*dt >= self.minNucleateDensity and Rcrit[p] >= self.Rmin[p]:
                Rad[0,p] = Rcrit[p] + 0.5 * np.sqrt(self.kB * T / (np.pi * self.gamma[p]))
            else:
                Rad[0,p] = 0

        self._currY[self.R_NUC] = Rad



import numpy as np
from kawin.precipitation.KWNBase import PrecipitateBase
from kawin.precipitation.PopulationBalance import PopulationBalanceModel
from kawin.precipitation.non_ideal.GrainBoundaries import GBFactors
from kawin.precipitation.Plot import plotEuler

class PrecipitateModel (PrecipitateBase):
    '''
    Euler implementation of KWN model

    Parameters
    ----------
    phases : list (optional)
        Precipitate phases (array of str)
        If only one phase is considered, the default is ['beta']
    elements : list (optional)
        Solute elements in system
        Note: order of elements must correspond to order of elements set in Thermodynamics module
        If binary system, then defualt is ['solute']
    '''
    def __init__(self, phases = ['beta'], elements = ['solute']):
        super().__init__(phases, elements)

        if self.numberOfElements == 1:
            self._growthRate = self._growthRateBinary
            self._Beta = self._BetaBinary1
        else:
            self._growthRate = self._growthRateMulti
            self._Beta = self._BetaMulti

        self.eqAspectRatio = [None for p in range(len(self.phases))]

    def _resetArrays(self):
        '''
        Resets and initializes arrays for all variables

        In addition to PrecipitateBase, the equilibrium aspect ratio area and population balance models are created here
        '''
        super()._resetArrays()
        self.PBM = [PopulationBalanceModel() for p in self.phases]

        self.RdrivingForceIndex = np.zeros(len(self.phases), dtype=np.int32)
        self.dissolutionIndex = np.zeros(len(self.phases), dtype=np.int32)
        
    def reset(self):
        '''
        Resets model results
        '''
        super().reset()

        for i in range(len(self.phases)):
            self.PBM[i].reset()
            self.PBM[i].resetRecordedData()

    def _addExtraSaveVariables(self, saveDict):
        for p in range(len(self.phases)):
            saveDict['PBM_data_' + self.phases[p]] = [self.PBM[p].min, self.PBM[p].max, self.PBM[p].bins]
            saveDict['PBM_PSD_' + self.phases[p]] = self.PBM[p].PSD
            saveDict['PBM_bounds_' + self.phases[p]] = self.PBM[p].PSDbounds
            saveDict['PBM_size_' + self.phases[p]] = self.PBM[p].PSDsize
            saveDict['eqAspectRatio_' + self.phases[p]] = self.eqAspectRatio[p]

    def load(filename):
        data = np.load(filename)
        model = PrecipitateModel(data['phases'], data['elements'])
        model._loadData(data)
        return model

    def _loadExtraVariables(self, data):
        for p in range(len(self.phases)):
            PBMdata = data['PBM_data_' + self.phases[p]]
            psd = data['PBM_PSD_' + self.phases[p]]
            bounds = data['PBM_bounds_' + self.phases[p]]
            size = data['PBM_size_' + self.phases[p]]
            eqAR = data['eqAspectRatio_' + self.phases[p]]
            self.PBM[p] = PopulationBalanceModel(PBMdata[0], PBMdata[1], int(PBMdata[2]))
            self.PBM[p].PSD = psd
            self.PBM[p].PSDsize = size
            self.PBM[p].PSDbounds = bounds
            self.eqAspectRatio[p] = eqAR

    def setPBMParameters(self, cMin = 1e-10, cMax = 1e-9, bins = 150, minBins = 100, maxBins = 200, adaptive = True, phase = None):
        '''
        Sets population balance model parameters for each phase

        Parameters
        ----------
        cMin : float
            Minimum bin size
        cMax : float
            Maximum bin size
        bins : int
            Initial number of bins
        minBins : int
            Minimum number of bins - will not be used if adaptive = False
        maxBins : int
            Maximum number of bins - will not be used if adaptive = False
        adaptive : bool
            Sets adaptive bin sizes - bins may still change upon nucleation
        phase : str
            Phase to consider (will set all phases if phase = None or 'all')
        '''
        if phase is None or phase == 'all':
            for p in range(len(self.phases)):
                self.PBM[p] = PopulationBalanceModel(cMin, cMax, bins, minBins, maxBins)
                self.PBM[p].setAdaptiveBinSize(adaptive)
        else:
            index = self.phaseIndex(phase)
            self.PBM[index] = PopulationBalanceModel(cMin, cMax, bins, minBins, maxBins)
            self.PBM[index].setAdaptiveBinSize(adaptive)

    def setPSDrecording(self, record = True, phase = 'all'):
        '''
        Sets recording parameters for PSD of specified phase

        Parameters
        ----------
        record : bool (optional)
            Whether to record PSD, defaults to True
        phase : str (optional)
            Precipitate phase to record for
            Defaults to 'all', which will apply to all precipitate phases
        '''
        if phase is None or phase == 'all':
            for p in self.phases:
                index = self.phaseIndex(p)
                self.PBM[index].setRecording(record)
        else:
            index = self.phaseIndex(phase)
            self.PBM[index].setRecording(record)

    def saveRecordedPSD(self, filename, compressed = True, phase = 'all'):
        '''
        Saves recorded PSD in npz format

        Parameters
        ----------
        filename : str
            File name to save to
            Note: the phase name will be added to the filename if all phases are being saved
        compressed : bool (optional)
            Whether to save in compressed npz format
            Defualts to True
        phase : str (optional)
            Phase to save PSD for
            Defaults to 'all', which will save a file for each phase
        '''
        if phase is None or phase == 'all':
            for p in self.phases:
                index = self.phaseIndex(p)
                self.PBM[index].saveRecordedPSD(filename + '_' + p, compressed)
        else:
            index = self.phaseIndex(phase)
            self.PBM[index].saveRecordedPSD(filename, compressed)

    def loadParticleSizeDistribution(self, data, phase = None):
        '''
        Loads particle size distribution for specified phase

        Parameters
        ----------
        data : array
            Array of data containing precipitate sizes
        phase : str (optional)
            Phase to consider (defaults to first precipitate in list)
        '''
        index = self.phaseIndex(phase)
        self.PBM[index].LoadDistribution(data)

    def particleRadius(self, phase = None):
        '''
        Returns PSD bounds of given phase

        Parameters
        ----------
        phase : str (optional)
            Phase to consider (defaults to first precipitate in list)
        '''
        index = self.phaseIndex(phase)
        return self.PBM[index].PSDbounds
        
    def particleGibbs(self, radius = None, phase = None):
        '''
        Returns Gibbs Thomson contribution of a particle given its radius
        
        Parameters
        ----------
        radius : array (optional)
            Precipitate radaii (defaults to None, which will use boundaries
                of the size classes of the precipitate PSD)
        phase : str (optional)
            Phase to consider (defaults to first precipitate in list)
        '''
        if radius is None:
            index = self.phaseIndex(phase)
            radius = self.PBM[index].PSDbounds
        return super().particleGibbs(radius, phase)

    def PSD(self, phase = None):
        '''
        Returns frequency of particle size distribution of given phase

        Parameters
        ----------
        phase : str (optional)
            Phase to consider (defaults to first precipitate in list)
        '''
        index = self.phaseIndex(phase)
        return self.PBM[index].PSD
    
    def createLookup(self):
        '''
        This creates a lookup table mapping the particle size classes to the interfacial composition
        '''
        #RdrivingForceIndex will find the index of the largest particle size class where the precipitate is unstable
        #This is determined by the interfacial composition function, where it should return -1 or None
        #All compositions from the PSD bounds will be set to the compositions just above RdrivingForceLimit
        #This is just to allow for particles to dissolve instead of pile up in the smallest bin
        self.RdrivingForceIndex = np.zeros(len(self.phases), dtype=np.int32)

        #Keep as separate arrays so that number of PSD classes can change within precipitate phases
        self.PSDXalpha = []
        self.PSDXbeta = []
        
        xEqAlpha = np.zeros((1, len(self.phases), self.numberOfElements))
        xEqBeta = np.zeros((1, len(self.phases), self.numberOfElements))
        T = self._currY[self.TEMPERATURE][0]
        for p in range(len(self.phases)):
            #Interfacial compositions at equilibrium (planar interface)
            xAResult, xBResult = self.interfacialComposition[p](T, 0)
            if xAResult == -1 or xAResult is None:
                xEqAlpha[0,p,0] = 0
                xEqBeta[0,p,0] = 0
            else:
                xEqAlpha[0,p,0] = xAResult
                xEqBeta[0,p,0] = xBResult

            #Interfacial compositions at each size class in PSD
            self.PSDXalpha.append(np.zeros((self.PBM[p].bins + 1, 1)))
            self.PSDXbeta.append(np.zeros((self.PBM[p].bins + 1, 1)))

            self.PSDXalpha[p][:,0], self.PSDXbeta[p][:,0] = self.interfacialComposition[p](T, self.particleGibbs(self.PBM[p].PSDbounds, self.phases[p]))
            self.RdrivingForceIndex[p] = np.argmax(self.PSDXalpha[p][:,0] != -1)-1
            self.RdrivingForceIndex[p] = 0 if self.RdrivingForceIndex[p] < 0 else self.RdrivingForceIndex[p]
            self.RdrivingForceLimit[p] = self.PBM[p].PSDbounds[self.RdrivingForceIndex[p]]

            #Sets particle radii smaller than driving force limit to driving force limit composition
            #If RdrivingForceIndex is at the end of the PSDX arrays, then no precipitate in the size classes of the PSD is stable
            #This can occur in non-isothermal situations where the temperature gets too high
            if self.RdrivingForceIndex[p]+1 < len(self.PSDXalpha[p][:,0]):
                self.PSDXalpha[p][:self.RdrivingForceIndex[p]+1,0] = self.PSDXalpha[p][self.RdrivingForceIndex[p]+1,0]
                self.PSDXbeta[p][:self.RdrivingForceIndex[p]+1,0] = self.PSDXbeta[p][self.RdrivingForceIndex[p]+1,0]
            else:
                self.PSDXalpha[p] = np.zeros((self.PBM[p].bins + 1,1))
                self.PSDXbeta[p] = np.zeros((self.PBM[p].bins + 1,1))

        return xEqAlpha, xEqBeta
            
    def setup(self):
        '''
        Sets up additional variables in addition to PrecipitateBase

        Sets up additional outputs, population balance models, equilibrium aspect ratio and equilibrium compositions
        '''
        if self._isSetup:
            return

        super().setup()

        #Equilibrium aspect ratio and PBM setup
        #If calculateAspectRatio is True, then use strain energy to calculate aspect ratio for each size class in PSD
        #Else, then use aspect ratio defined in shape factors
        self.eqAspectRatio = [None for p in range(len(self.phases))]
        for p in range(len(self.phases)):
            self.PBM[p].reset()

            if self.calculateAspectRatio[p]:
                self.eqAspectRatio[p] = self.strainEnergy[p].eqAR_bySearch(self.PBM[p].PSDbounds, self.gamma[p], self.shapeFactors[p])
                arFunc = lambda R, p1=p : self._interpolateAspectRatio(R, p1)
                self.shapeFactors[p].setAspectRatio(arFunc)
            else:
                self.eqAspectRatio[p] = self.shapeFactors[p].aspectRatio(self.PBM[p].PSDbounds)

        self._currY = [np.array([self.varList[i][self.n]]) for i in range(self.NUM_TERMS)]
        self._currY[self.TIME] = np.array([self.time[self.n]])
        self._currY[self.TEMPERATURE] = np.array([self.getTemperature(self.time[self.n])])
        
        #Setup interfacial composition
        if self.numberOfElements == 1:
            self.xEqAlpha[self.n], self.xEqBeta[self.n] = self.createLookup()
        else:
            self.PSDXalpha = [None for p in range(len(self.phases))]
            self.PSDXbeta = [None for p in range(len(self.phases))]

            #Set first index of eq composition
            for p in range(len(self.phases)):
                #Use arbitrary dg, R and gE since only the eq compositions are needed here
                _, _, _, xEqAlpha, xEqBeta = self.interfacialComposition[p](self.xComp[self.n], self.temperature[self.n], 0, 1, 0)
                if xEqAlpha is not None:
                    self.xEqAlpha[self.n,p] = xEqAlpha
                    self.xEqBeta[self.n,p] = xEqBeta
        
        x = [self.PBM[p].PSD for p in range(len(self.phases))]
        self._calcDrivingForce(self.time[self.n], x)
        self._growthRate()
        self._calcNucleationRate(self.time[self.n], x)
        for i in range(self.NUM_TERMS):
            self.varList[i][self.n] = self._currY[i][0]
    
    def _interpolateAspectRatio(self, R, p):
        '''
        Linear interpolation between self.eqAspectRatio and self.PBM[p].PSDbounds

        Parameters
        ----------
        R : float
            Equivalent spherical radius
        p : int
            Phase index
        '''
        return np.interp(R, self.PBM[p].PSDbounds, self.eqAspectRatio[p])

    def getDt(self, dXdt):
        '''
        The following checks are made
            1) change in number of particles moving between bins
                This is controlled by the implementation in PopulationBalanceModel,
                but essentially limits the number of particles moving between bins
            2) change in nucleation rate
                Time will be proportional to the 1/log(previous nuc rate / new nuc rate)
            3) change in temperature
                Limits how fast temperature can change
            4) change in critical radius
                Proportional to a percent change in critical radius
            5) estimated change in volume fraction
                Estimates the change in volume fraction from the nucleation rate and nucleation radius
        '''
        #Start test dt at 0.01 or previous dt
        i = self.n
        dtPrev = 0.01 if self.n == 0 else self.time[i] - self.time[i-1]
        #Try to slowly increase the time step
        #  Precipitation kinetics is more on a log scale than linear (unless temperature changes are involve)
        #  Thus, we can get away with increasing the time step over time assuming that kinetics are slowing down
        #  Plus, unlike the single phase diffusion module, there's no form way to define a good time step apart from the checks here
        dtPropose = (1 + self.dtScale) * dtPrev
        dtMax = self.finalTime - self.time[i]
        
        dtAll = [dtMax]
        if self.checkPSD:
            dtPBM = [dtMax]
            if i > 0 and self.temperature[i] == self.temperature[i-1]:
                dtPBM += [self.PBM[p].getDTEuler(dtMax, self.growth[p], self.dissolutionIndex[p]) for p in range(len(self.phases))]
            dtPBM = np.amin(dtPBM)
            dtAll.append(dtPBM)
        
        if self.checkNucleation:
            dtNuc = dtMax * np.ones(len(self.phases))
            if i > 0:
                nRateCurr = self.nucRate[i]
                nRatePrev = self.nucRate[i-1]
                for p in range(len(self.phases)):
                    if nRateCurr[p] > self.minNucleationRate and nRatePrev[p] > self.minNucleationRate and nRatePrev[p] != nRateCurr[p]:
                        dtNuc[p] = self.maxNucleationRateChange * dtPrev / np.abs(np.log10(nRatePrev[p] / nRateCurr[p]))
            else:
                for p in range(len(self.phases)):
                    if self.nucRate[i,p] * dtPrev > 1e5:
                        dtNuc[p] = 1e5 / self.nucRate[i,p]
            dtNuc = np.amin(dtNuc)
            dtAll.append(dtNuc)

        #Temperature change constraint
        if self.checkTemperature and i > 0:
            Tchange = self.temperature[i] - self.temperature[i-1]
            dtTemp = dtMax
            if Tchange > self.maxNonIsothermalDT:
                dtTemp = self.maxNonIsothermalDT * dtPrev / Tchange
            dtAll.append(dtTemp)

        if self.checkRcrit and i > 0:
            dtRad = dtMax * np.ones(len(self.phases))
            if not all((self.Rcrit[i-1,:] == 0) & (self.Rcrit[i,:] - self.Rcrit[i-1,:] == 0) & (self.dGs[i,:] <= 0)):
                indices = (self.Rcrit[i-1,:] > 0) & (self.Rcrit[i,:] - self.Rcrit[i-1,:] != 0) & (self.dGs[i,:] > 0)
                dtRad[indices] = self.maxRcritChange * dtPrev / np.abs((self.Rcrit[i,:][indices] - self.Rcrit[i-1,:][indices]) / self.Rcrit[i-1,:][indices])
            dtRad = np.amin(dtRad)
            dtAll.append(dtRad)

        if self.checkVolumePre:
            dV = np.zeros(len(self.phases))
            for p in range(len(self.phases)):
                #Calculate estimate volume change based off growth rate and nucleated particles
                #TODO: account for non-spherical precipitates
                dVi = self.PBM[p].PSD * self.PBM[p].PSDsize**2 * 0.5 * (self.growth[p][1:] + self.growth[p][:-1])
                dVi[dVi < 0] = 0
                dV = self.VmAlpha / self.VmBeta[p] * (self.GB[p].areaFactor * np.sum(dVi) + self.GB[p].volumeFactor * self.nucRate[i,p] * self.Rad[i,p]**3)

            dtVol = dtMax * np.ones(len(self.phases))
            for p in range(len(self.phases)):
                if dV != 0:
                    dtVol[p] = self.maxVolumeChange / (2 * np.abs(dV))
            dtVol = np.amin(dtVol)
            dtAll.append(dtVol)

        dt = np.amin(dtAll)
        #If all time checks pass, then go back to previous time step and increase it slowly
        #   This is so we don't step at the maximum possible time
        if dt == dtMax:
            dt = dtPropose

        return dt
    
    def _processX(self, x):
        '''
        Quick check to make sure particles below the thresholds are 0
            RdrivingForceIndex - only for binary, where energy from the Gibbs-Thompson effect is high enough
                that the free energy of the precipitate is above the free energy surface of the matrix phase
                and equilibrium cannot be calculated
            minRadius - minimum radius to be considered a precipitate
        '''
        for p in range(len(self.phases)):
            x[p][:self.RdrivingForceIndex[p]+1] = 0
            x[p][self.PBM[p].PSDsize < self.minRadius] = 0
        return
    
    def _calcNucleationRate(self, t, x):
        '''
        The _calcNucleationRate function in KWNBase calculates the nucleation rate as the
            probability that a site can form a nucleate that will continue to grow

        To convert this probability to an actual nucleation rate, we multiply by the amount
            of available nucleation sites

        The number of available sites is determined by:
            Available sites = All sites - used up sites + sites on parent precipitates
            The used up sites depends on the type of nucleation
                Bulk and grain corners - used sites = number of current precipitates
                Dislocation and grain edges - number of sites filled along the edges (assumes average radius of precipitates)
                Grain boundaries - number of sites filled along the faces (assumes average cross sectional area of precipitates)
        '''
        super()._calcNucleationRate(t, x)
        for p in range(len(self.phases)):
            #If parent phases exists, then calculate the number of potential nucleation sites on the parent phase
            #This is the number of lattice sites on the total surface area of the parent precipitate
            nucleationSites = np.sum([4 * np.pi * self.PBM[p2].SecondMomentFromN(x[p2]) * (self.avo / self.VmBeta[p2])**(2/3) for p2 in self.parentPhases[p]])

            if self.GB[p].nucleationSiteType == GBFactors.BULK:
                #bulkPrec = np.sum([self.GB[p2].volumeFactor * self.PBM[p2].ThirdMoment() for p2 in range(len(self.phases)) if self.GB[p2].nucleationSiteType == GBFactors.BULK])
                #nucleationSites += self.bulkN0 - bulkPrec * (self.avo / self.VmAlpha)
                bulkPrec = np.sum([self.PBM[p2].ZeroMomentFromN(x[p2]) for p2 in range(len(self.phases)) if self.GB[p2].nucleationSiteType == GBFactors.BULK])
                nucleationSites += self.bulkN0 - bulkPrec
            elif self.GB[p].nucleationSiteType == GBFactors.DISLOCATION:
                bulkPrec = np.sum([self.PBM[p2].FirstMomentFromN(x[p2]) for p2 in range(len(self.phases)) if self.GB[p2].nucleationSiteType == GBFactors.DISLOCATION])
                nucleationSites += self.dislocationN0 - bulkPrec * (self.avo / self.VmAlpha)**(1/3)
            elif self.GB[p].nucleationSiteType == GBFactors.GRAIN_BOUNDARIES:
                boundPrec = np.sum([self.GB[p2].gbRemoval * self.PBM[p2].SecondMomentFromN(x[p2]) for p2 in range(len(self.phases)) if self.GB[p2].nucleationSiteType == GBFactors.GRAIN_BOUNDARIES])
                nucleationSites += self.GBareaN0 - boundPrec * (self.avo / self.VmAlpha)**(2/3)
            elif self.GB[p].nucleationSiteType == GBFactors.GRAIN_EDGES:
                edgePrec = np.sum([np.sqrt(1 - self.GB[p2].GBk**2) * self.PBM[p2].FirstMomentFromN(x[p2]) for p2 in range(len(self.phases)) if self.GB[p2].nucleationSiteType == GBFactors.GRAIN_EDGES])
                nucleationSites += self.GBedgeN0 - edgePrec * (self.avo / self.VmAlpha)**(1/3)
            elif self.GB[p].nucleationSiteType == GBFactors.GRAIN_CORNERS:
                cornerPrec = np.sum([self.PBM[p2].ZeroMomentFromN(x[p2]) for p2 in range(len(self.phases)) if self.GB[p2].nucleationSiteType == GBFactors.GRAIN_CORNERS])
                nucleationSites += self.GBcornerN0 - cornerPrec
               
            if nucleationSites < 0:
                nucleationSites = 0
            self._currY[self.NUC_RATE][0,p] *= nucleationSites
    
    def _calcMassBalance(self, t, x):
        '''
        Mass balance to find matrix composition with new particle size distribution

        This also includes: volume fraction, precipitate density, average radius, average aspect ratio and sum of precipitate composition
        '''
        fBeta = np.zeros((1,len(self.phases)))
        fConc = np.zeros((1, len(self.phases),self.numberOfElements))
        precDens = np.zeros((1,len(self.phases)))
        avgR = np.zeros((1,len(self.phases)))
        avgAR = np.zeros((1,len(self.phases)))
        xComp = np.zeros((1,self.numberOfElements))
        
        for p in range(len(self.phases)):
            volRatio = self.VmAlpha / self.VmBeta[p]
            Ntot = self.PBM[p].ZeroMomentFromN(x[p])
            #If no precipitates, then avgR, avgAR, precDens, fConc and fBeta for phase p is all 0
            if Ntot == 0:
                continue
            RadSum = self.PBM[p].MomentFromN(x[p], 1)
            ARsum = self.PBM[p].WeightedMomentFromN(x[p], 0, self.shapeFactors[p].aspectRatio(self.PBM[p].PSDsize))
            fBeta[0,p] = np.amin([volRatio * self.GB[p].volumeFactor * self.PBM[p].ThirdMomentFromN(x[p]), 1])

            '''
            Concentration of the precipitates - needed to get matrix composition

            For a line compound with composition x^beta, this boils down to:
            x_0 = (1-f_v) * x^inf + f_v * x^beta
                Where x_0 is initial composition, f_v is volume fraction and x^inf is matrix composition

            For non-stoichiometric compounds, we want to integrate the precipitate composition as a function of radius
                We'll call this term f_conc (fraction + concentration of precipitates), so:
                x_0 = (1-f_v) * x^inf + f_conc
            
            For infinite precipitate diffusion, the concentration of a single precipitate is assumed to be homogenous
            f_conc = r_vol * vol_factor * sum(n_i * R_i^3 * x_i^beta)
                Where r_vol is V^alpha / V^beta and vol_factor is a factor for converting R^3 to volume (for sphere, this is 4*pi/3)

            For no diffusion in precipitate, the concentration depends on the history of the precipitate compositions and growth rate
            We just have to convert the summation to an integral of the time derivative of the terms inside
            f_conc = r_vol * vol_factor * sum(int(d(n_i * R_i^3 * x_i^beta)/dt, dt))
            We'll assume x_i^beta is constant with time (for 3 or more components, this is not true, but assume it doesn't change significantly per iteration - it'll also be a lot harder to account for)
            d(f_conc)/dt = r_vol * vol_factor * sum(d(n_i)/dt * R_i^3 * x_i^beta + 3 * R_i^3 * d(R_i)/dt * n_i * x_i^beta)
                d(n_i)/dt is the change in precipitates, since we don't record this, this is just (x[p] - self.PBM[p].PSD) / dt - with x[p] being the new number density for phase p
                d(R_i)/dt is the growth rate, however, since we use a eulerian solver, this corresponds to the growth rate of the bins themselves, which is 0
                    If we were to use a langrangian solver, then d(n_i)/dt would be 0 (since the density in each bin would be constant) and d(R_i)/dt would be the growth rate at R_i
            Then we can calculate f_conc per iteration as a time integration like we do with some of the other variables
            '''
            if self.infinitePrecipitateDiffusion[p]:
                compAvg = 0.5 * (self.PSDXbeta[p][:-1] + self.PSDXbeta[p][1:])
                for e in range(self.numberOfElements):
                    fConc[0,p,e] = volRatio * self.GB[p].volumeFactor * self.PBM[p].WeightedMomentFromN(x[p], 3, compAvg[:,e])
            else:
                midX = (self.PSDXbeta[p][1:] + self.PSDXbeta[p][:-1]) / 2
                for e in range(self.numberOfElements):
                    #y = volRatio * self.GB[p].volumeFactor * np.sum((3*midG*self.PBM[p].PSDsize**2*self.PBM[p].PSD*dt + self.PBM[p].PSDsize**3*(x[p]-self.PBM[p].PSD))*midX[:,e])
                    y = volRatio * self.GB[p].volumeFactor * np.sum((self.PBM[p].PSDsize**3*(x[p]-self.PBM[p].PSD))*midX[:,e])
                    fConc[0,p,e] = self.fConc[self.n,p,e] + y

            #Only record these terms if there are non-zero number of precipitates
            #Otherwise we will be dividing by 0 for avgR and avgAR
            #   Argueably, RadSum and ARsum would be 0 if Ntot is 0, so it should be fine to do this
            if Ntot > self.minNucleateDensity:
                avgR[0,p] = RadSum / Ntot
                precDens[0,p] = Ntot
                avgAR[0,p] = ARsum / Ntot
            else:
                avgR[0,p] = 0
                precDens[0,p] = 0
                avgAR[0,p] = 0

            #Not sure if needed, but just in case
            if self.betaFrac[self.n,p] == 1:
                fBeta[0,p] = 1

        if np.sum(fBeta[0]) < 1:
            xComp[0] = (self.xComp[0] - np.sum(fConc[0], axis=0)) / (1 - np.sum(fBeta[0]))
            xComp[0,xComp[0] < 0] = self.minComposition

        self._currY[self.VOL_FRAC] = fBeta
        self._currY[self.FCONC] = fConc
        self._currY[self.PREC_DENS] = precDens
        self._currY[self.R_AVG] = avgR
        self._currY[self.AR_AVG] = avgAR
        self._currY[self.COMPOSITION] = xComp

    def getCurrentX(self):
        '''
        Returns current value of time and X
        In this case, X is the particle size distribution for each phase
        '''
        return self.time[self.n], [self.PBM[p].PSD for p in range(len(self.phases))]

    def _getdXdt(self, t, x):
        '''
        Returns dn_i/dt for each PBM of each phase
        '''
        return [self.PBM[p].getdXdtEuler(self.growth[p], self._currY[self.NUC_RATE][0,p], self._currY[self.R_NUC][0,p], x[p]) for p in range(len(self.phases))]

    def correctdXdt(self, dt, x, dXdt):
        '''
        Corrects dXdt with the newly found dt, this adjusts the fluxes at the ends of the PBM so that we don't get negative bins
        '''
        for p in range(len(self.phases)):
            dXdt[p] = self.PBM[p].correctdXdtEuler(dt, self.growth[p], self._currY[self.NUC_RATE][0,p], self._currY[self.R_NUC][0,p], x[p])

    def _singleGrowthBinary(self, p):
        '''
        Calculates growth rate for a single phase
        This is separated from _growthRateBinary since it's used in _calculatePSD

        Matrix/precipitate composition are not calculated here since it's
        already calculated in createLookup
        '''
        xComp = self._currY[self.COMPOSITION][0]
        T = self._currY[self.TEMPERATURE][0]
        growthRate = np.zeros(self.PBM[p].bins + 1)
        #If no precipitates are stable, don't calculate growth rate and set PSD to 0
        #This should represent dissolution of the precipitates
        if self.RdrivingForceIndex[p]+1 < len(self.PSDXalpha[p][:,0]):
            superSaturation = (xComp[0] - self.PSDXalpha[p][:,0]) / (self.VmAlpha * self.PSDXbeta[p][:,0] / self.VmBeta[p] - self.PSDXalpha[p][:,0])
            growthRate = self.shapeFactors[p].kineticFactor(self.PBM[p].PSDbounds) * self.Diffusivity(xComp[0], T) * superSaturation / (self.effDiffDistance(superSaturation) * self.PBM[p].PSDbounds)

        return growthRate
    
    def _growthRateBinary(self):
        '''
        Determines current growth rate of all particle size classes in a binary system
        '''
        #Update equilibrium interfacial compositions
        #This will be override if createLookup is called
        T = self._currY[self.TEMPERATURE]
        self.dTemp += T - self.temperature[self.n]
        if np.abs(self.dTemp) > self.maxTempChange:
            xEqAlpha, xEqBeta = self.createLookup()
        else:
            xEqAlpha, xEqBeta = np.array([self.xEqAlpha[self.n]]), np.array([self.xEqBeta[self.n]])
            self.dTemp = 0
        self._currY[self.EQ_COMP_ALPHA] = xEqAlpha
        self._currY[self.EQ_COMP_BETA] = xEqBeta
        
        #growthRate = np.zeros((len(self.phases), self.bins + 1))
        growthRate = []
        for p in range(len(self.phases)):
            growthRate.append(self._singleGrowthBinary(p))
            
        self.growth = growthRate

    def _singleGrowthMulti(self, p):
        '''
        Calculates growth rate for a single phase
        This is separated from _growthRateMulti since it's used in _calculatePSD

        This will also calculate the matrix/precipitate composition 
        for the radius in the PSD as well as equilibrium (infinite radius)
        '''
        xComp = self._currY[self.COMPOSITION][0]
        dGs = self._currY[self.DRIVING_FORCE][0]
        T = self._currY[self.TEMPERATURE][0]
        precDens = self._currY[self.PREC_DENS][0]
        if dGs[p] < 0 and precDens[p] <= 0:
            xEqAlpha = np.zeros(self.numberOfElements)
            xEqBeta = np.zeros(self.numberOfElements)
            growthRate = np.zeros(self.PBM[p].bins + 1)
            return growthRate, xEqAlpha, xEqBeta


        growth, xAlpha, xBeta, xEqAlpha, xEqBeta = self.interfacialComposition[p](xComp, T, dGs[p] * self.VmBeta[p], self.PBM[p].PSDbounds, self.particleGibbs(phase=self.phases[p]), searchDir = self._precBetaTemp[p])

        #If two-phase equilibrium not found, two possibilities - precipitates are unstable or equilibrium calculations didn't converge
        #We try to avoid this as much as possible to where if precipitates are unstable, then attempt to get a growth rate from the nearest composition on the phase boundary
        #And if equilibrium calculations didn't converge, try to use the previous calculations assuming the new composition is close to the previous
        if growth is None:
            #If driving force is negative, then precipitates are unstable
            if dGs[p] < 0:
                #Completely reset the PBM, including bounds and number of bins
                #In case nucleation occurs again, the PBM will be at a good length scale
                self.PSDXalpha[p] = np.zeros((self.PBM[p].bins + 1, self.numberOfElements))
                self.PSDXbeta[p] = np.zeros((self.PBM[p].bins + 1, self.numberOfElements))
                xEqAlpha = np.zeros(self.numberOfElements)
                xEqBeta = np.zeros(self.numberOfElements)
                growthRate = np.zeros(self.PBM[p].bins + 1)
            #Else, equilibrium did not converge and just use previous values
            #Only the growth rate needs to be updated, since all other terms are previous
            #Also revert the PSD in case this function was called to adjust for the new PSD bins
            else:
                growthRate = self.growth[p]
        else:
            #Update interfacial composition for each precipitate size
            self.PSDXalpha[p] = xAlpha
            self.PSDXbeta[p] = xBeta

            #Add shape factor to growth rate - will need to add effective diffusion distance as well
            growthRate = self.shapeFactors[p].kineticFactor(self.PBM[p].PSDbounds) * growth

        return growthRate, xEqAlpha, xEqBeta
    
    def _growthRateMulti(self):
        '''
        Determines current growth rate of all particle size classes in a multicomponent system
        '''
        xEqAlpha = np.zeros((1,len(self.phases), self.numberOfElements))
        xEqBeta = np.zeros((1,len(self.phases), self.numberOfElements))
        growthRate = []
        for p in range(len(self.phases)):
            growthRate_p, xEqAlpha_p, xEqBeta_p = self._singleGrowthMulti(p)
            growthRate.append(growthRate_p)
            xEqAlpha[0,p] = xEqAlpha_p
            xEqBeta[0,p] = xEqBeta_p
        self._currY[self.EQ_COMP_ALPHA] = xEqAlpha
        self._currY[self.EQ_COMP_BETA] = xEqBeta
        self.growth = growthRate

    def _updateParticleSizeDistribution(self, t, x):
        '''
        Updates particle size distribution with new x

        Steps:
            1. Check if growth rate calculation failed with negative driving force
                We'll reset the PBM since we can't do much from here, but the chances of this happening should be pretty low
            2. Update the PBM with new x
            3. Check if the PBM needs to adjust the size class
                If so, then update the cached aspect ratio and precipitate composition with the new size classes
            4. Remove precipitates below a certain threshold (RdrivingForceIndex and minRadius)
            5. Calculate the dissolution index (index at which below are not considered when calculating dt)
                This is to prevent very small dt as the growth rate increases rapidly when R->0
        '''
        for p in range(len(self.phases)):
            if self.dGs[self.n,p] < 0 and np.all(self.xEqAlpha[self.n,p,:] == 0):
                self.PBM[p].reset()
                self.PSDXalpha[p] = np.zeros((self.PBM[p].bins + 1, self.numberOfElements))
                self.PSDXbeta[p] = np.zeros((self.PBM[p].bins + 1, self.numberOfElements))
                self.growth[p] = np.zeros(self.PBM[p].bins+1)
                continue
            self.PBM[p].UpdatePBMEuler(t, x[p])
            change, addedIndices = self.PBM[p].adjustSizeClassesEuler(all(self.growth[p] < 0))
            if change:
                if self.calculateAspectRatio[p]:
                    self.eqAspectRatio[p] = self.strainEnergy[p].eqAR_bySearch(self.PBM[p].PSDbounds, self.gamma[p], self.shapeFactors[p])
                else:
                    self.eqAspectRatio[p] = self.shapeFactors[p].aspectRatio(self.PBM[p].PSDbounds)

                self.growth[p] = np.zeros(len(self.PBM[p].PSDbounds))
                if self.numberOfElements == 1:
                    if addedIndices is None:
                        #This is very slow to do
                        self.createLookup()
                    else:
                        self.PSDXalpha[p] = np.concatenate((self.PSDXalpha[p], np.zeros((self.PBM[p].bins+1 - len(self.PSDXalpha[p]),1))))
                        self.PSDXbeta[p] = np.concatenate((self.PSDXbeta[p], np.zeros((self.PBM[p].bins+1 - len(self.PSDXbeta[p]),1))))
                        self.PSDXalpha[p][addedIndices:,0], self.PSDXbeta[p][addedIndices:,0] = self.interfacialComposition[p](self.temperature[self.n], self.particleGibbs(self.PBM[p].PSDbounds[addedIndices:], self.phases[p]))
                else:
                    self.PSDXalpha[p] = np.zeros((self.PBM[p].bins + 1, self.numberOfElements))
                    self.PSDXbeta[p] = np.zeros((self.PBM[p].bins + 1, self.numberOfElements))
                self._growthRate()
            self.PBM[p].PSD[:self.RdrivingForceIndex[p]+1] = 0
            self.PBM[p].PSD[self.PBM[p].PSDsize < self.minRadius] = 0
            self.dissolutionIndex[p] = self.PBM[p].getDissolutionIndex(self.maxDissolution, self.RdrivingForceIndex[p])
            #self.PBM[p].PSD[:self.dissolutionIndex[p]] = 0

    def plot(self, axes, variable, bounds = None, timeUnits = 's', radius='spherical', *args, **kwargs):
        '''
        Plots model outputs
        
        Parameters
        ----------
        axes : Axis
        variable : str
            Specified variable to plot
            Options are 'Volume Fraction', 'Total Volume Fraction', 'Critical Radius',
                'Average Radius', 'Volume Average Radius', 'Total Average Radius', 
                'Total Volume Average Radius', 'Aspect Ratio', 'Total Aspect Ratio'
                'Driving Force', 'Nucleation Rate', 'Total Nucleation Rate',
                'Precipitate Density', 'Total Precipitate Density', 
                'Temperature', 'Composition',
                'Size Distribution', 'Size Distribution Curve',
                'Size Distribution KDE', 'Size Distribution Density
                'Interfacial Composition Alpha', 'Interfacial Composition Beta'

                Note: for multi-phase simulations, adding the word 'Total' will
                    sum the variable for all phases. Without the word 'Total', the variable
                    for each phase will be plotted separately

                    Interfacial composition terms are more relavent for binary systems than
                    for multicomponent systems
                    
        bounds : tuple (optional)
            Limits on the x-axis (float, float) or None (default, this will set bounds to (initial time, final time))
        radius : str (optional)
            For non-spherical precipitates, plot the Average Radius by the -
                Equivalent spherical radius ('spherical')
                Short axis ('short')
                Long axis ('long')
            Note: Total Average Radius and Volume Average Radius will still use the equivalent spherical radius
        *args, **kwargs - extra arguments for plotting
        '''
        plotEuler(self, axes, variable, bounds, timeUnits, radius, *args, **kwargs)


                


from tinydb import where
import numpy as np
from pycalphad import Model, variables as v
from pycalphad.core.utils import wrap_symbol, extract_parameters
from pycalphad.io.tdb import get_supported_variables
from symengine import exp, Symbol, Add
from kawin.thermo.FreeEnergyHessian import partialdMudX, dMudX

#List of interstitial elements
# When calculating interdiffusivity, we do not require reference element
# When calculating the mobility factor, we have an additional vacancy term to multiply by
#As a list here, hopefully this should be editable by a user outside of this module - may have to edit __init__.py
interstitials = ['C', 'N', 'O', 'H', 'B']

class MobilityModel(Model):
    '''
    Handles mobility and diffusivity data from .tdb files

    Parameters
    ----------
    dbe : Database
    comps : list
        List of elements to consider
    phase_name : str
    parameters : dict or list (optional)
        Dictionary of parameters to be kept symbolic

    Attributes
    ----------
    mobility_variables : dict
        Dictionary of symbols in mobility functions for each element
    diffusivity_variables : dict
        Dictionary of symbols in diffusivity functions for each element
    '''
    def __init__(self, dbe, comps, phase_name, parameters=None):
        super().__init__(dbe, comps, phase_name, parameters)

        symbols = {Symbol(s): val for s, val in dbe.symbols.items()}
        if self._parameters_arg is not None:
            if isinstance(self._parameters_arg, dict):
                symbols.update([(wrap_symbol(s), val) for s, val in self._parameters_arg.items()])
            else:
                # Lists of symbols that should remain symbolic
                for s in self._parameters_arg:
                    symbols.pop(wrap_symbol(s))

        #Replace symbols with database symbols for mobility and exponential term
        #Also store a copy of the mobility/diffusivity as MOB_A or DIFF_A
        for name, value in self.mobility.items():
            self.mobility[name] = self.symbol_replace(value, symbols).xreplace(get_supported_variables())
            setattr(self, 'MOB_'+name, self.mobility[name])
            setattr(self, 'MQ_'+name, self.symbol_replace(getattr(self, 'MQ_'+name), symbols).xreplace(get_supported_variables()))

        for name, value in self.diffusivity.items():
            self.diffusivity[name] = self.symbol_replace(value, symbols).xreplace(get_supported_variables())
            setattr(self, 'DIFF_'+name, self.diffusivity[name])
            setattr(self, 'DQ_'+name, self.symbol_replace(getattr(self, 'DQ_'+name), symbols).xreplace(get_supported_variables()))

        self.mob_site_fractions = {c: sorted([x for x in self.mobility_variables[c] if isinstance(x, v.SiteFraction)], key=str) for c in self.mobility}
        self.diff_site_fractions = {c: sorted([x for x in self.diffusivity_variables[c] if isinstance(x, v.SiteFraction)], key=str) for c in self.diffusivity}
        self.mob_state_variables = {c: sorted([x for x in self.mobility_variables[c] if not isinstance(x, v.SiteFraction)], key=str) for c in self.mobility}
        self.diff_state_variables = {c: sorted([x for x in self.diffusivity_variables[c] if not isinstance(x, v.SiteFraction)], key=str) for c in self.diffusivity}

    @property
    def mobility_variables(self):
        '''
        List of variables in the mobility functions

        Returns
        -------
        dictionary {component name (str) : variables (list)}
        '''
        return {c: sorted([x for x in self.mobility[c].free_symbols if isinstance(x, v.StateVariable)], key=str) for c in self.mobility}

    @property
    def diffusivity_variables(self):
        '''
        List of variables in the diffusivity functions

        Returns
        -------
        dictionary {component name (str) : variables (list)}
        '''
        return {c: sorted([x for x in self.diffusivity[c].free_symbols if isinstance(x, v.StateVariable)], key=str) for c in self.diffusivity}

    def build_phase(self, dbe):
        '''
        Builds free energy and mobility/diffusivity models as abstract syntax tree

        Parameters
        ----------
        dbe : Database
        '''
        super().build_phase(dbe)
        self.mobility, self.diffusivity = self.build_mobility(dbe)

    def _mobility_validity(self, constituent_array):
        '''
        Returns true if constituent_array contains only active species of current model
        Ignores vacancies - if a sublattice is only vacancies, then it should be ignored
            For now, phases where this occurs will have the parameters listed for n-1 sublattices
        '''
        for param_sublattice, model_sublattice in zip(constituent_array, self.constituents):
            if not (set(param_sublattice).issubset(model_sublattice) or (param_sublattice[0] == v.Species('*'))):
                return False
        return True
        
    def build_mobility(self, dbe):
        '''
        Builds mobility/diffusivity models as abstract syntax tree

        Parameters
        ----------
        dbe : Database
        '''
        phase = dbe.phases[self.phase_name]
        param_search = dbe.search

        self.mob_models = {}
        mob = {}
        diff = {}

        param_names = ['MF', 'MQ', 'DF', 'DQ']

        for c in self.components:
            if c.name != 'VA':
                for name in param_names:
                    param_query = (
                        (where('phase_name') == phase.name) & \
                        (where('parameter_type') == name) & \
                        (where('constituent_array').test(self._mobility_validity)) & \
                        (where('diffusing_species') == c)
                    )
                    rk = self.redlich_kister_sum(phase, param_search, param_query)

                    if name not in self.mob_models:
                        self.mob_models[name] = {}
                    self.mob_models[name][c.name] = rk

                #Additional parameters search if diffusing species are not included
                #   This is mainly intended to help with parameter fitting
                #   Parameters will be in the format for MOB_A, MOB_B (or the respective keyword)
                #   This will reflect how the models are stored in MobilityModel
                #       Ex. MOB_A is the entire mobility model of A while MQ_A is the redlich kister polynomial used for A mobility
                #Additional parameters will be in tuples of (database keyword, mob_models keyword)
                additional_params = [('MOB', 'MQ'), ('MQ', 'MQ'), ('DIFF', 'DQ'), ('DQ', 'DQ')]
                for p in additional_params:
                    fit_name = p[0] + '_' + c.name
                    param_query = (
                        (where('phase_name') == phase.name) & \
                        (where('parameter_type') == fit_name) & \
                        (where('constituent_array').test(self._mobility_validity))
                    )
                    rk = self.redlich_kister_sum(phase, param_search, param_query)
                    self.mob_models[p[1]][c.name] += rk

        self.checkOrderingContribution(dbe)
        for c in self.components:
            if c.name != 'VA':
                #In thermo-calc, the mobility model is defined as exp(sum(MF)/RT) * exp(sum(MQ)/RT) / RT
                #The diffusivity model is defined either as dilute - exp(sum(DF)/RT) * exp(sum(DQ)/RT)
                #                                        or simple - sum(DF) + sum(DQ)
                # We use the dilute assumption here
                #In summary, there's no difference between MF and MQ, or between DF and DQ
                # For papers using Q and theta (pre-exponential term), corrections must be made to theta have it fit the definitions above
                mqsum = self.mob_models['MF'][c.name] + self.mob_models['MQ'][c.name]
                dqsum = self.mob_models['DF'][c.name] + self.mob_models['DQ'][c.name]
                mob[c.name] = (1 / (v.R * v.T)) * exp(mqsum / (v.R * v.T))
                diff[c.name] = exp(dqsum / (v.R * v.T))

                #Also store the exponential term in case we want to grab the activation energy or pre-exp term
                setattr(self, 'MQ_'+str(c.name).upper(), mqsum)
                setattr(self, 'DQ_'+str(c.name).upper(), dqsum)

        return mob, diff
    
    def checkOrderingContribution(self, dbe):
        '''
        Checks if phase is an ordered part of a order-disorder model

        The ordered part of the phase double counts the disordered contribution, so the model is
        G = G_dis + G_ord(y) - G_ord(y=x)

        This is straight up copied from Model.atomic_ordering_energy in pycalphad with
          the minor difference that we replace the symbols in mob and diff
        '''
        phase = dbe.phases[self.phase_name]
        ordered_phase_name = phase.model_hints.get('ordered_phase', None)
        disordered_phase_name = phase.model_hints.get('disordered_phase', None)

        #If not order-disorder model, then return as unchanged
        if phase.name != ordered_phase_name:
            return
        
        ordered_phase = dbe.phases[ordered_phase_name]
        constituents = [sorted(set(c).intersection(self.components)) for c in ordered_phase.constituents]
        disordered_phase = dbe.phases[disordered_phase_name]
        disordered_model = self.__class__(dbe, sorted(self.components), disordered_phase_name)

        disordered_subl_constituents = disordered_phase.constituents[0]
        ordered_constituents = ordered_phase.constituents
        substitutional_sublattice_idxs = []
        for idx, subl_constituents in enumerate(ordered_constituents):
            if len(disordered_subl_constituents.symmetric_difference(subl_constituents)) == 0:
                substitutional_sublattice_idxs.append(idx)

        num_substitutional_sublattice_idxs = len(substitutional_sublattice_idxs)
        num_ordered_interstitial_subls = len(ordered_phase.sublattices) - num_substitutional_sublattice_idxs
        num_disordered_interstitial_subls = len(disordered_phase.sublattices) - 1
        if num_ordered_interstitial_subls != num_disordered_interstitial_subls:
            raise ValueError(
                f'Number of interstitial sublattices for the disordered phase '
                f'({num_disordered_interstitial_subls}) and the ordered phase '
                f'({num_ordered_interstitial_subls}) do not match. Got '
                f'substitutional sublattice indices of {substitutional_sublattice_idxs}.'
                )
        
        for c in self.mob_models['MF']:
            ordered_mobQ = Add(self.mob_models['MQ'][c])
            ordered_mobF = Add(self.mob_models['MF'][c])
            ordered_diffQ = Add(self.mob_models['DQ'][c])
            ordered_diffF = Add(self.mob_models['DF'][c])

            # Compute the molefraction_dict, which will map ordered phase site
            # fractions to the quasi mole fractions representing the disordered state
            molefraction_dict = {}
            ordered_sitefracs = [x for x in ordered_mobQ.free_symbols if isinstance(x, v.SiteFraction)]
            for sitefrac in ordered_sitefracs:
                if sitefrac.sublattice_index in substitutional_sublattice_idxs:
                    molefraction_dict[sitefrac] = \
                        self._quasi_mole_fraction(sitefrac.species,
                                                ordered_phase_name,
                                                constituents,
                                                ordered_phase.sublattices,
                                                substitutional_sublattice_idxs,
                                                )

            # Compute the variable_rename_dict, which will map disordered phase site
            # fractions to the quasi mole fractions representing the disordered state
            variable_rename_dict = {}
            disordered_sitefracs = [x for x in disordered_model.energy.free_symbols if isinstance(x, v.SiteFraction)]
            for atom in disordered_sitefracs:
                if atom.sublattice_index == 0:  # only the first sublattice is substitutional
                    variable_rename_dict[atom] = \
                        self._quasi_mole_fraction(atom.species,
                                                ordered_phase_name,
                                                constituents,
                                                ordered_phase.sublattices,
                                                substitutional_sublattice_idxs,
                                                )

                else:
                    shifted_subl_index = atom.sublattice_index + num_substitutional_sublattice_idxs - 1
                    variable_rename_dict[atom] = \
                        v.SiteFraction(ordered_phase_name, shifted_subl_index, atom.species)
                    
            self.mob_models['MQ'][c] = self._partitioned_expr(disordered_model.mob_models['MQ'][c], ordered_mobQ, variable_rename_dict, molefraction_dict)
            self.mob_models['MF'][c] = self._partitioned_expr(disordered_model.mob_models['MF'][c], ordered_mobF, variable_rename_dict, molefraction_dict)
            self.mob_models['DQ'][c] = self._partitioned_expr(disordered_model.mob_models['DQ'][c], ordered_diffQ, variable_rename_dict, molefraction_dict)
            self.mob_models['DF'][c] = self._partitioned_expr(disordered_model.mob_models['DF'][c], ordered_diffF, variable_rename_dict, molefraction_dict)

        return

def mobility_from_composition_set(composition_set, mobility_callables = None, mobility_correction = None, parameters = {}):
    '''
    Computes mobility from equilibrium results

    Parameters
    ----------
    composition_set : pycalphad.core.composition_set.CompositionSet
    mobility_callables : dict
        Pre-computed mobility callables for each element
    mobility_correction : dict (optional)
        Factor to multiply mobility by for each given element (defaults to 1)
    parameters : dict {str : float}
        List of parameters to override free symbols in the model

    Returns
    -------
    Array for floats for mobility of each element (alphabetical order)
    '''
    if mobility_callables is None:
        raise ValueError('mobility_callables is required')

    elements = list(composition_set.phase_record.nonvacant_elements)

    #Set mobility correction if not set
    if mobility_correction is None:
        mobility_correction = {A: 1 for A in elements}
    else:
        for A in elements:
            if A not in mobility_correction:
                mobility_correction[A] = 1

    #return np.array([mobility_correction[elements[A]] * mobility_callables[elements[A]](composition_set.dof) for A in range(len(elements))])
    param_keys, param_values = extract_parameters(parameters)
    if len(param_values) > 0:
        callableInput = np.concatenate((composition_set.dof, param_values[0]), dtype=np.float64)
    else:
        callableInput = composition_set.dof
    return np.array([mobility_correction[elements[A]] * mobility_callables[elements[A]](callableInput) for A in range(len(elements))])
    
def tracer_diffusivity(composition_set, mobility_callables = None, mobility_correction = None, parameters = {}):
    '''
    Computes tracer diffusivity for given equilibrium results
    D = MRT

    Parameters
    ----------
    composition_set : pycalphad.core.composition_set.CompositionSet
    mobility_callables : dict
        Pre-computed mobility callables for each element
    mobility_correction : dict (optional)
        Factor to multiply mobility by for each given element (defaults to 1)

    Returns
    -------
    Array of floats of diffusivity for each element (alphabetical order)
    '''
    if mobility_callables is None:
        raise ValueError('mobility_callables is required')

    R = 8.314
    T = composition_set.dof[composition_set.phase_record.state_variables.index(v.T)]
    return R * T * mobility_from_composition_set(composition_set, mobility_callables, mobility_correction, parameters)

def mobility_matrix(composition_set, mobility_callables = None, mobility_correction = None, parameters = {}):
    '''
    Mobility matrix
    Used to obtain diffusivity when multipled with free energy hessian

    Parameters
    ----------
    composition_set : pycalphad.core.composition_set.CompositionSet
    mobility_callables : dict (optional)
        Pre-computed mobility callables for each element
    mobility_correction : dict (optional)
        Factor to multiply mobility by for each given element (defaults to 1)

    Returns
    -------
    Matrix of floats
        Each index along an axis correspond to elements in alphabetical order
    '''
    elements = list(composition_set.phase_record.nonvacant_elements)
    X = composition_set.X

    #U-fraction - defined as U_a = X_a / sum(substitutionals)
    Usum = np.sum([X[A] for A in range(len(elements)) if elements[A] not in interstitials])
    U = X / Usum

    #Multiply mobility by U-fraction for ease of use when constructing the mobility matrix
    computedMob = mobility_from_composition_set(composition_set, mobility_callables, mobility_correction, parameters)
    mob = np.array([U[A] * computedMob[A] for A in range(len(elements))])

    #Find vacancy site fractions for multiplying with interstitials when making the mobility matrix
    #If vacancies are not found on the same sublattice, we'll defualt to 1 so there's at least some mobility and not 0
    #       A mobility of 0 would be quite unrealistic
    #       In addition, as we're working with interstitals, the vacancies are going to be close to 1, so this assumption wouldn't hurt
    vaTerms = {}            #Maps sublattice index to site fraction index for vacancies
    interstitialTerms = {}  #Maps interstitial to sublattice index
    index = len(composition_set.phase_record.state_variables)
    for i in range(len(composition_set.phase_record.variables)):
        if composition_set.phase_record.variables[i].species.name == 'VA':
            vaTerms[composition_set.phase_record.variables[i].sublattice_index] = composition_set.dof[index+i]
        if composition_set.phase_record.variables[i].species.name in interstitials:
            interstitialTerms[composition_set.phase_record.variables[i].species.name] = composition_set.phase_record.variables[i].sublattice_index

    #For interstitials
    #    M_aa = y_Va * M_a
    #    y_Va is taken from the same sublattice that a is on, where more vacancies on the sublattice implies faster diffusion
    #For substitutionals
    #    M_aa = (1-U_a) * U_a * M_a
    #    M_ab = -U_a * U_b * M_b
    #There are no entries for M_ab if one index is interstitial and the other is substitutional
    mobMatrix = np.zeros((len(elements), len(elements)))
    for a in range(len(elements)):
        if elements[a] in interstitials:
            mobMatrix[a, a] = vaTerms.get(interstitialTerms[elements[a]], 1) * mob[a]
        else:
            for b in range(len(elements)):
                if elements[b] not in interstitials:
                    if a == b:
                        mobMatrix[a, b] = (1 - U[a]) * mob[b]
                    else:
                        mobMatrix[a, b] = -U[a] * mob[b]
    #Diffusivity requires dmu_a/dU_b; however, the free energy curvature gives dmu_a/dX_b
    #Assuming that Usum is constant and using chain-rule derivatives,
    #    the conversion from dmu_a/dX_b to dmu_a/dU_b can be done by multiplying the sum(substitutionals)
    mobMatrix *= Usum

    #Old way of computing mobility assuming only substitutional elements (so much simpler...)
    #mob = np.array([X[A] * computedMob[A] for A in range(len(elements))])
    #for a in range(len(elements)):
    #    for b in range(len(elements)):
    #        if a == b:
    #            mobMatrix[a, b] = (1 - X[a]) * mob[b]
    #        else:
    #            mobMatrix[a, b] = -X[a] * mob[b]

    return mobMatrix

def chemical_diffusivity(chemical_potentials, composition_set, mobility_callables, mobility_correction = None, returnHessian = False, parameters = {}):
    '''
    Chemical diffusivity (D_kj)
        D_kj = sum((delta_ik - U_k) * U_i * M_i) * dmu_i/dU_j
        D_ab = mobility matrix * free energy hessian

    Parameters
    ----------
    chemical_potentials : 1-D ndarray
    composition_set : pycalphad.core.composition_set.CompositionSet
    mobility_callables : dict
        Pre-computed mobility callables for each element
    mobility_correction : dict (optional)
        Factor to multiply mobility by for each given element (defaults to 1)
    returnHessian : bool (optional)
        Whether to return chemical potential derivative (defaults to False)

    Returns
    -------
    (matrix of floats, free energy hessian)
        Each index along an axis correspond to elements in alphabetical order
        free energy hessian will be None if returnHessian is False
    '''
    dmudx = partialdMudX(chemical_potentials, composition_set)
    #print('dmudx', dmudx)
    mobMatrix = mobility_matrix(composition_set, mobility_callables, mobility_correction, parameters)
    #print('mobMatrix', mobMatrix)
    Dkj = np.matmul(mobMatrix, dmudx)
    
    if returnHessian:
        return Dkj, dmudx
    else:
        return Dkj, None

def interdiffusivity(chemical_potentials, composition_set, refElement, mobility_callables = None, mobility_correction = None, returnHessian = False, parameters = {}):
    '''
    Interdiffusivity (D^n_ab)

    D^n_ab = D_ab - D_an (for substitutional element)
    D^n_ab = D_ab (for interstitial element)

    Parameters
    ----------
    chemical_potentials : 1-D ndarray
    composition_set : pycalphad.core.composition_set.CompositionSet
    refElement : str
        Reference element n
    mobility_callables : dict (optional)
        Pre-computed mobility callables for each element
    mobility_correction : dict (optional)
        Factor to multiply mobility by for each given element (defaults to 1)
    returnHessian : bool (optional)
        Whether to return chemical potential derivative (defaults to False)

    Returns
    -------
    (matrix of floats, free energy hessian)
        Each index along an axis correspond to elements in 
            alphabetical order excluding reference element
        free energy hessian will be None if returnHessian is False
    '''
    Dkj, hessian = chemical_diffusivity(chemical_potentials, composition_set, mobility_callables, mobility_correction, returnHessian, parameters)
    #print('Dkj', Dkj)
    elements = list(composition_set.phase_record.nonvacant_elements)

    #Find index of reference element
    refIndex = 0
    for a in range(len(elements)):
        if elements[a] == refElement:
            refIndex = a
            break

    #Build Dnkj, skipping the reference element
    Dnkj = np.zeros((len(elements) - 1, len(elements) - 1))
    c = 0
    d = 0
    for a in range(len(elements)):
        if a != refIndex:
            for b in range(len(elements)):
                if b != refIndex:
                    if elements[b] in interstitials:
                        Dnkj[c, d] = Dkj[a, b]
                    else:
                        Dnkj[c, d] = Dkj[a, b] - Dkj[a, refIndex]
                    d += 1
            c += 1
            d = 0

    return Dnkj, hessian

def inverseMobility(chemical_potentials, composition_set, refElement, mobility_callables, mobility_correction = None, returnOther = True, parameters = {}):
    '''
    Inverse mobility matrix for determining interfacial composition from 
        Philippe and P. W. Voorhees, Acta Materialia 61 (2013) p. 4237

    M^-1 = (free energy hessian) * Dnkj^-1

    Parameters
    ----------
    chemical_potentials : 1-D ndarray
    composition_set : pycalphad.core.composition_set.CompositionSet
    refElement : str
        Reference element n
    mobility_callables : dict
        Pre-computed mobility callables for each element
    mobility_correction : dict (optional)
        Factor to multiply mobility by for each given element (defaults to 1)
    returnOther : bool (optional)
        Whether to return interdiffusivity and hessian (defaults to False)

    Returns
    -------
    (interdiffusivity, hessian, inverse mobility)
        Interdiffusivity and hessian will be None if returnOther is False
    '''
    Dnkj, _ = interdiffusivity(chemical_potentials, composition_set, refElement, mobility_callables, mobility_correction, False, parameters)
    totalH = dMudX(chemical_potentials, composition_set, refElement)
    #print('totalH', totalH)
    if returnOther:
        return Dnkj, totalH, np.matmul(totalH, np.linalg.inv(Dnkj))
    else:
        return None, None, np.matmul(totalH, np.linalg.inv(Dnkj))
    
def tracer_diffusivity_from_diff(composition_set, diffusivity_callables = None, diffusivity_correction = None, parameters = {}):
    '''
    Tracer diffusivity from diffusivity callables

    This will just return the Da as an array

    Parameters
    ----------
    composition_set : pycalphad.core.composition_set.CompositionSet
    diffusivity_callables : dict
        Pre-computed diffusivity callables for each element
    diffusivity_correction : dict (optional)
        Factor to multiply diffusivity by for each given element (defaults to 1)

    Returns
    -------
    Array of floats of diffusivity for each element (alphabetical order)
    '''
    if diffusivity_callables is None:
        raise ValueError('diffusivity_callables is required')

    elements = list(composition_set.phase_record.nonvacant_elements)

    #Set diffusivity correction if not set
    if diffusivity_correction is None:
        diffusivity_correction = {A: 1 for A in elements}
    else:
        for A in elements:
            if A not in diffusivity_correction:
                diffusivity_correction[A] = 1

    #return np.array([diffusivity_correction[elements[A]] * diffusivity_callables[elements[A]](composition_set.dof) for A in range(len(elements))])

    param_keys, param_values = extract_parameters(parameters)
    if len(param_values) > 0:
        callableInput = np.concatenate((composition_set.dof, param_values[0]), dtype=np.float64)
    else:
        callableInput = composition_set.dof
    return np.array([diffusivity_correction[elements[A]] * diffusivity_callables[elements[A]](callableInput) for A in range(len(elements))])

def interdiffusivity_from_diff(composition_set, refElement, diffusivity_callables, diffusivity_correction = None, parameters = {}):
    '''
    Interdiffusivity (D^n_ab) calculated from diffusivity callables
        This is if the TDB database only has diffusivity data and no mobility data

    D^n_ab = D_ab - D_an (for substitutional element)
    D^n_ab = D_ab (for interstitial element)

    Parameters
    ----------
    composition_set : pycalphad.core.composition_set.CompositionSet
    refElement : str
        Reference element n
    diffusivity_callables : dict
        Pre-computed callables for diffusivity for each element
    diffusivity_correction : dict
        Factor to multiply diffusivity by for each element (defaults to 1)

    Returns
    -------
    matrix of floats
        Each index along an axis correspond to elements in 
            alphabetical order excluding reference element
    '''
    elements = list(composition_set.phase_record.nonvacant_elements)

    if diffusivity_correction is None:
        diffusivity_correction = {elements[A]: 1 for A in elements}
    else:
        for A in elements:
            if A not in diffusivity_correction:
                diffusivity_correction[A] = 1

    param_keys, param_values = extract_parameters(parameters)
    if len(param_values) > 0:
        callableInput = np.concatenate((composition_set.dof, param_values[0]), dtype=np.float64)
    else:
        callableInput = composition_set.dof
    Dnkj = np.zeros((len(elements) - 1, len(elements) - 1))
    eleIndex = 0
    for a in range(len(elements) - 1):
        if elements[eleIndex] == refElement:
            eleIndex += 1
            
        #Daa = diffusivity_correction[elements[eleIndex]] * diffusivity_callables[elements[eleIndex]](composition_set.dof)
        Daa = diffusivity_correction[elements[eleIndex]] * diffusivity_callables[elements[eleIndex]](callableInput)
        Dnkj[a, a] = Daa

        eleIndex += 1

    return Dnkj

def inverseMobility_from_diffusivity(chemical_potentials, composition_set, refElement, diffusivity_callables, diffusivity_correction = None, returnOther = True, parameters = {}):
    '''
    Inverse mobility matrix for determining interfacial composition

    M^-1 = (free energy hessian) * Dnkj^-1

    Parameters
    ----------
    chemical_potentials : 1-D ndarray
    composition_set : pycalphad.core.composition_set.CompositionSet
    refElement : str
        Reference element n
    diffusivity_callables : dict
        Pre-computed callables for diffusivity for each element
    diffusivity_correction : dict
        Factor to multiply diffusivity by for each element (defaults to 1)
    returnOther : bool (optional)
        Whether to return interdiffusivity and hessian (defaults to False)

    Returns
    -------
    (interdiffusivity, hessian, inverse mobility)
        Interdiffusivity and hessian will be None if returnOther is False
    '''
    Dnkj = interdiffusivity_from_diff(composition_set, refElement, diffusivity_callables, diffusivity_correction, parameters)
    totalH = dMudX(chemical_potentials, composition_set, refElement)

    if returnOther:
        return Dnkj, totalH, np.matmul(totalH, np.linalg.inv(Dnkj))
    else:
        return None, None, np.matmul(totalH, np.linalg.inv(Dnkj))





import numpy as np
from kawin.diffusion.Diffusion import DiffusionModel
from kawin.thermo.Mobility import mobility_from_composition_set
import copy

class HomogenizationModel(DiffusionModel):
    def __init__(self, zlim, N, elements = ['A', 'B'], phases = ['alpha'], record = True):
        super().__init__(zlim, N, elements, phases, record)

        self.mobilityFunction = self.wienerUpper
        self.defaultMob = 0
        self.eps = 0.05

        self.sortIndices = np.argsort(self.allElements)
        self.unsortIndices = np.argsort(self.sortIndices)
        self.labFactor = 1

    def reset(self):
        '''
        Resets model

        This also includes chemical potential and pycalphad CompositionSets for each node
        '''
        super().reset()
        self.mu = np.zeros((len(self.elements)+1, self.N))
        self.compSets = [None for _ in range(self.N)]

    def setMobilityFunction(self, function):
        '''
        Sets averaging function to use for mobility

        Default mobility value should be that a phase of unknown mobility will be ignored for average mobility calcs

        Parameters
        ----------
        function : str
            Options - 'upper wiener', 'lower wiener', 'upper hashin-shtrikman', 'lower hashin-strikman', 'labyrinth'
        '''
        #np.finfo(dtype).max - largest representable value
        #np.finfo(dtype).tiny - smallest positive usable value
        if 'upper' in function and 'wiener' in function:
            self.mobilityFunction = self.wienerUpper
            self.defaultMob = np.finfo(np.float64).tiny
        elif 'lower' in function and 'wiener' in function:
            self.mobilityFunction = self.wienerLower
            self.defaultMob = np.finfo(np.float64).max
        elif 'upper' in function and 'hashin' in function:
            self.mobilityFunction = self.hashin_shtrikmanUpper
            self.defaultMob = np.finfo(np.float64).tiny
        elif 'lower' in function and 'hashin' in function:
            self.mobilityFunction = self.hashin_shtrikmanLower
            self.defaultMob = np.finfo(np.float64).max
        elif 'lab' in function:
            self.mobilityFunction = self.labyrinth
            self.defaultMob = np.finfo(np.float64).tiny

    def setLabyrinthFactor(self, n):
        '''
        Labyrinth factor

        Parameters
        ----------
        n : int
            Either 1 or 2
            Note: n = 1 will the same as the weiner upper bounds
        '''
        if n < 1:
            n = 1
        if n > 2:
            n = 2
        self.labFactor = n

    def setup(self):
        '''
        Sets up model

        This also includes getting the CompositionSets for each node
        '''
        super().setup()
        #self.midX = 0.5 * (self.x[:,1:] + self.x[:,:-1])
        self.p = self.updateCompSets(self.x)

    def _newEqCalc(self, x, T):
        '''
        Calculates equilibrium and returns a CompositionSet
        '''
        wks = self.therm.getEq(x, T, 0, self.phases)
        chemical_potentials = np.squeeze(wks.eq.MU)
        composition_sets = wks.get_composition_sets()
        return chemical_potentials, composition_sets

    def updateCompSets(self, xarray):
        '''
        Updates the array of CompositionSets

        If an equilibrium calculation is already done for a given composition, 
        the CompositionSet will be taken out of the hash table

        Otherwise, a new equilibrium calculation will be performed

        Parameters
        ----------
        xarray : (e-1, N) array
            Composition for each node
            e is number of elements
            N is number of nodes

        Returns
        -------
        parray : (p, N) array
            Phase fractions for each node
            p is number of phases
        '''
        parray = np.zeros((len(self.phases), xarray.shape[1]))
        for i in range(parray.shape[1]):
            if self.cache:
                hashValue = self._getHash(xarray[:,i], self.T[i])
                if hashValue not in self.hashTable:
                    chemical_potentials, comp = self._newEqCalc(xarray[:,i], self.T[i])
                    self.hashTable[hashValue] = (chemical_potentials, comp, None)
                else:
                    chemical_potentials, comp, _ = self.hashTable[hashValue]
                chemical_potentials, self.compSets[i] = copy.copy(chemical_potentials), copy.copy(comp)
            else:
                chemical_potentials, self.compSets[i] = self._newEqCalc(xarray[:,i], self.T[i])
            self.mu[:,i] = chemical_potentials[self.unsortIndices]
            cs_phases = [cs.phase_record.phase_name for cs in self.compSets[i]]
            for p in range(len(cs_phases)):
                parray[self._getPhaseIndex(cs_phases[p]), i] = self.compSets[i][p].NP
        
        return parray

    def getMobility(self, xarray):
        '''
        Gets mobility of all phases

        Returns
        -------
        (p, e+1, N) array - p is number of phases, e is number of elements, N is number of nodes
        '''
        mob = self.defaultMob * np.ones((len(self.phases), len(self.elements)+1, xarray.shape[1]))
        for i in range(xarray.shape[1]):
            if self.cache:
                hashValue = self._getHash(xarray[:,i], self.T[i])
                _, _, mTemp = self.hashTable[hashValue]
            else:
                mTemp = None
            if mTemp is None or not self.cache:
                maxPhaseAmount = 0
                maxPhaseIndex = 0
                for p in range(len(self.phases)):
                    if self.p[p,i] > 0:
                        if self.p[p,i] > maxPhaseAmount:
                            maxPhaseAmount = self.p[p,i]
                            maxPhaseIndex = p
                        if self.phases[p] in self.therm.mobCallables and self.therm.mobCallables[self.phases[p]] is not None:
                            #print(self.phases, self.phases[p], xarray[:,i], self.p[:,i], i, self.compSets[i])
                            compset = [cs for cs in self.compSets[i] if cs.phase_record.phase_name == self.phases[p]][0]
                            mob[p,:,i] = mobility_from_composition_set(compset, self.therm.mobCallables[self.phases[p]], self.therm.mobility_correction)[self.unsortIndices]
                            mob[p,:,i] *= np.concatenate(([1-np.sum(xarray[:,i])], xarray[:,i]))
                        else:
                            mob[p,:,i] = -1
                for p in range(len(self.phases)):
                    if any(mob[p,:,i] == -1) and not all(mob[p,:,i] == -1):
                        mob[p,:,i] = mob[maxPhaseIndex,:,i]
                    if all(mob[p,:,i] == -1):
                        mob[p,:,i] = self.defaultMob
                if self.cache:
                    self.hashTable[hashValue] = (self.hashTable[hashValue][0], self.hashTable[hashValue][1], copy.copy(mob[:,:,i]))
            else:
                mob[:,:,i] = mTemp

        return mob

    def wienerUpper(self, xarray):
        '''
        Upper wiener bounds for average mobility

        Returns
        -------
        (e+1, N) mobility array - e is number of elements, N is number of nodes
        '''
        mob = self.getMobility(xarray)
        avgMob = np.sum(np.multiply(self.p[:,np.newaxis], mob), axis=0)
        return avgMob

    def wienerLower(self, xarray):
        '''
        Lower wiener bounds for average mobility

        Returns
        -------
        (e+1, N) mobility array - e is number of elements, N is number of nodes
        '''
        #(p, e, N)
        mob = self.getMobility(xarray)
        avgMob = 1/np.sum(np.multiply(self.p[:,np.newaxis], 1/mob), axis=0)
        return avgMob

    def labyrinth(self, xarray):
        '''
        Labyrinth mobility

        Returns
        -------
        (e+1, N) mobility array - e is number of elements, N is number of nodes
        '''
        mob = self.getMobility(xarray)
        avgMob = np.sum(np.multiply(np.power(self.p[:,np.newaxis], self.labFactor), mob), axis=0)
        return avgMob

    def hashin_shtrikmanUpper(self, xarray):
        '''
        Upper hashin shtrikman bounds for average mobility

        Returns
        -------
        (e+1, N) mobility array - e is number of elements, N is number of nodes
        '''
        #self.p                                 #(p,N)
        mob = self.getMobility(xarray)          #(p,e+1,N)
        maxMob = np.amax(mob, axis=0)           #(e+1,N)

        # 1 / ((1 / mPhi - mAlpha) + 1 / (3mAlpha)) = 3mAlpha * (mPhi - mAlpha) / (2mAlpha + mPhi)
        Ak = 3 * maxMob * (mob - maxMob) / (2*maxMob + mob)
        Ak = Ak * self.p[:,np.newaxis]
        Ak = np.sum(Ak, axis=0)
        avgMob = maxMob + Ak / (1 - Ak / (3*maxMob))
        return avgMob

    def hashin_shtrikmanLower(self, xarray):
        '''
        Lower hashin shtrikman bounds for average mobility

        Returns
        -------
        (e, N) mobility array - e is number of elements, N is number of nodes
        '''
        #self.p                                 #(p,N)
        mob = self.getMobility(xarray)          #(p,e+1,N)
        minMob = np.amin(mob, axis=0)           #(e+1,N)

        #This prevents an infinite mobility which could cause the time interval to be 0
        minMob[minMob == np.inf] = 0

        # 1 / ((1 / mPhi - mAlpha) + 1 / (3mAlpha)) = 3mAlpha * (mPhi - mAlpha) / (2mAlpha + mPhi)
        Ak = 3 * minMob * (mob - minMob) / (2*minMob + mob)

        Ak = Ak * self.p[:,np.newaxis]
        Ak = np.sum(Ak, axis=0)
        avgMob = minMob + Ak / (1 - Ak / (3*minMob))
        return avgMob
    
    def _getFluxes(self, t, x_curr):
        '''
        Return fluxes and time interval for the current iteration

        Steps:
            1. Get average mobility from homogenization function. Interpolate to get mobility (M) at cell boundaries
            2. Interpolate composition to get composition (x) at cell boundaries
            3. Calculate chemical potential gradient (dmu/dz) at cell boundaries
            4. Calculate composition gradient (dx/dz) at cell boundaries
            5. Calculate homogenization flux = -M / dmu/dz
            6. Calculate ideal contribution = -eps * M*R*T / x * dx/dz
            7. Apply boundary conditions for fluxes at ends of mesh
                If fixed flux condition (Neumann) - then use the flux defined in the condition
                If fixed composition condition (Dirichlet) - then use nearby flux (this will keep the composition fixed after apply the fluxes)

        TODO: If using RK4, I believe the phase fraction will be from the last step of the RK4 iteration. May not make sense to do that
        '''
        x = x_curr[0]
        self.T = self.Tfunc(self.z, t)
        self.p = self.updateCompSets(x)

        #Get average mobility between nodes
        avgMob = self.mobilityFunction(x)
        avgMob = 0.5 * (avgMob[:,1:] + avgMob[:,:-1])

        #Composition between nodes
        avgX = 0.5 * (x[:,1:] + x[:,:-1])
        avgX = np.concatenate(([1-np.sum(avgX, axis=0)], avgX), axis=0)

        #Chemical potential gradient
        dmudz = (self.mu[:,1:] - self.mu[:,:-1]) / self.dz

        #Composition gradient (we need to calculate gradient for reference element)
        dxdz = (x[:,1:] - x[:,:-1]) / self.dz
        dxdz = np.concatenate(([0-np.sum(dxdz, axis=0)], dxdz), axis=0)

        # J = -M * dmu/dz
        # Ideal contribution: J_id = -eps * M*R*T / x * dx/dz
        fluxes = np.zeros((len(self.elements)+1, self.N-1))
        fluxes = -avgMob * dmudz
        nonzeroComp = avgX != 0
        Tmid = (self.T[1:] + self.T[:-1]) / 2
        Tmidfull = Tmid[np.newaxis,:]
        for i in range(fluxes.shape[0]-1):
            Tmidfull = np.concatenate((Tmidfull, Tmid[np.newaxis,:]), axis=0)
        fluxes[nonzeroComp] += -self.eps * avgMob[nonzeroComp] * 8.314 * Tmidfull[nonzeroComp] * dxdz[nonzeroComp] / avgX[nonzeroComp]

        #Flux in a volume fixed frame: J_vi = J_i - x_i * sum(J_j)
        vfluxes = np.zeros((len(self.elements), self.N+1))
        vfluxes[:,1:-1] = fluxes[1:,:] - avgX[1:,:] * np.sum(fluxes, axis=0)

        #Boundary conditions
        for e in range(len(self.elements)):
            vfluxes[e,0] = self.LBCvalue[e] if self.LBC[e] == self.FLUX else vfluxes[e,1]
            vfluxes[e,-1] = self.RBCvalue[e] if self.RBC[e] == self.FLUX else vfluxes[e,-2]

        return vfluxes

    def getFluxes(self):
        '''
        Return fluxes and time interval for the current iteration
        '''
        vfluxes = self._getFluxes(self.t, [self.x])
        dJ = np.abs(vfluxes[:,1:] - vfluxes[:,:-1]) / self.dz
        dt = self.maxCompositionChange / np.amax(dJ[dJ!=0])
        return vfluxes, dt
    
    def getDt(self, dXdt):
        '''
        Time increment
        This is done by finding the time interval such that the composition
            change caused by the fluxes will be lower than self.maxCompositionChange
        '''
        return self.maxCompositionChange / np.amax(np.abs(dXdt[0][dXdt[0]!=0]))